#include "mex.h"
#include "csources/svm.h"
#include "csources/svm_utils.h"
#include "csources/svm_params.h"
#include "csources/utils.h"
#include "csources/DiskUtils.h"
#include "csources/kernel_func.h"
#include "csources/Data.h"
#include "csources/MXData.h"
#include "csources/grid_search.h"
#include "csources/cross_validation.h"

#include <cstring>

using namespace std;

void mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
	PRINTF("nlhs: %i, nrhs: %i\n", nlhs, nrhs);

	if ((nlhs >= 8) && (nrhs >= 2)) {
	
		const mxArray* mx_X 		    = prhs[0];
		const mxArray* mx_c 		    = prhs[1];
		const mxArray* mx_k 		    = prhs[2];
		const mxArray* mx_kernel_type       = prhs[3];
		const mxArray* mx_grid_count        = (nrhs > 4) ? prhs[4] : NULL;
		const mxArray* mx_precompute_kernel = (nrhs > 5) ? prhs[5] : NULL;
		const mxArray* mx_max_level         = (nrhs > 6) ? prhs[6] : NULL;
		const mxArray* mx_KERNEL_CACHE_MB   = (nrhs > 7) ? prhs[7] : NULL;
		const mxArray* mx_save_memory       = (nrhs > 8) ? prhs[8] : NULL;

		int return_best_mean_error    = (nlhs > 8) ? 1 : 0;
		int return_best_mean_sv_count = (nlhs > 9) ? 1 : 0;
		int return_C                  = (nlhs > 10) ? 1 : 0;

		int save_memory = (mx_save_memory != NULL)  ? ((!mxIsEmpty(mx_save_memory)) ? *mxGetPr(mx_save_memory) : 0)  : 0;
		ClassInfoDataDesc* X = MXDataDesc::Create(mx_X, mx_c, save_memory);

		int X_rows    = X->ssize();
		int X_columns = X->size();
	
	        double*	c = double_array(X_columns);
		X->getClassInfo(c);		

		SVMconfig* config = new SVMconfig();
		config->output_off();

		double kernel_type = (mx_kernel_type != NULL)  ? ((!mxIsEmpty(mx_kernel_type)) ? *mxGetPr(mx_kernel_type) : KERNEL_RBF)  : KERNEL_RBF;
		config->set("kernel_type", kernel_type, X->size());

//		config->set("epsilon", 1e-1, X->size());
//
		double KERNEL_CACHE_MB = (mx_KERNEL_CACHE_MB != NULL)  ? ((!mxIsEmpty(mx_KERNEL_CACHE_MB )) ? *mxGetPr(mx_KERNEL_CACHE_MB ) : 500)  : 500;
		config->set("KERNEL_CACHE_MB", KERNEL_CACHE_MB, X->size());

		Kernel* K = Kernel::create_kernel(config->kernel_type());

		if (K==NULL) {
			PRINTF("Could not interpret kernel parameters ... bailing out.\n");
			return;
		}

		int precompute_K = (mx_precompute_kernel != NULL)  ? ((!mxIsEmpty(mx_precompute_kernel)) ? (int) *mxGetPr(mx_precompute_kernel) : 0)  : 0;	
		if (precompute_K) K->precompute(X);

		K->ComputeNormFactor(X);

		InfoConfig* iconfig = new InfoConfig(config->info_mode());	

		HashConfig* hconfig = new HashConfig();

		int k = (mx_k != NULL)  ? ((!mxIsEmpty(mx_k)) ? (int) *mxGetPr(mx_k) : 10)  : 10;

		int nparams              = K->nParams() + 1;
		const char** param_names = (const char**) pointer_array(nparams);
		param_names[0]           = "C";
		memcpy(param_names + 1, K->ParamNames(), K->nParams()*sizeof(char*));

		CrossValidation* cv = new CrossValidation(k, X, param_names, nparams, K, config, hconfig, iconfig);

		int max_level = (mx_max_level != NULL)  ? ((!mxIsEmpty(mx_max_level)) ? (int) *mxGetPr(mx_max_level) : 5)  : 5;	

		double* tmin      = double_array(nparams);
		double* tmax      = double_array(nparams);
		double* tcount    = double_array(nparams);
		double* tstop_tol = double_array(nparams);

		double grid_count = (mx_grid_count != NULL)  ? ((!mxIsEmpty(mx_grid_count)) ? *mxGetPr(mx_grid_count) : 5)  : 5;	

		assign(tcount, grid_count, nparams);

		for (int i=0; i < nparams; i++) {
			tmin[i] = K->GridMin(param_names[i], tmin[i]); 
			tmin[i] = config->GridMin(param_names[i], tmin[i]); 

			tmax[i] = K->GridMax(param_names[i], tmax[i]); 
			tmax[i] = config->GridMax(param_names[i], tmax[i]); 

			tstop_tol[i] = K->GridStopTol(param_names[i], tstop_tol[i], tcount[i]); 
			tstop_tol[i] = config->GridStopTol(param_names[i], tstop_tol[i], tcount[i]); 
		}

		double zoom_factor = 0.7;

		PRINTF("Number of training samples: %i, number of data dimensions: %i\n", X->size(), X->ssize());
		PRINTF("Performing %i-fold cross validation, grid count: %i, max refinement levels: %i...\n", k, (int) grid_count, max_level);
		PRINTF("Classifier base configuration: %s\n", config->ident_str());FLUSH;

		GridSearch* gs = new GridSearch(tmin, tmax, tcount, tstop_tol, zoom_factor, nparams, max_level, cv);

		double* best = double_array(nparams);
		
		double* best_func_value = gs->search(best);

	        double best_C = config->GridMap(param_names[0],best[0]);

		free_array(tmin);
		free_array(tmax);
		free_array(tcount);
		free_array(tstop_tol);

		for (int i=0; i < nparams; i++) {
			best[i] = K->GridMap(param_names[i],best[i]);
			best[i] = config->GridMap(param_names[i],best[i]);

			K->Set(param_names[i], best[i]);
			config->set(param_names[i], best[i], X->size());
		}

		double best_mean_error   = best_func_value[0];
		double best_mean_svcount = best_func_value[1];

		delete gs;

		PRINTF("best mean error: %g\n", best_mean_error);	
		PRINTF("best mean sv count: %g\n", best_mean_svcount);	
		print_array("best parameters: ", best, nparams);

		double **ALPHA = NULL;
		double* B      = NULL;	
		double** UCL   = NULL;
		int ccount     = 0;
		int UCL_rows   = 0;
		SubsetClassInfoDataDesc* SVs = NULL;

		config->output_on();

		SubsetClassInfoDataDesc* X_train = cv->GetBalancedSet();

 		if (!svm_wrapper (&ALPHA, &B, &SVs, &UCL, &UCL_rows, &ccount, X,  config, K, hconfig, iconfig)) {

			for (int i=0; i < nlhs; i++) plhs[i] = NULL;
		}
		else {
			int svCount = SVs->size();

			PRINTF("ccount: %i,  svCount: %i\n", ccount, svCount);

			double** RETURN_ALPHA = m_transpose(ALPHA, ccount, svCount);
			mxArray* mx_ALPHA = mxCreateNumericMatrix (ccount, svCount, mxDOUBLE_CLASS, mxREAL);	
			m_to_vec(mxGetPr(mx_ALPHA), RETURN_ALPHA, svCount, ccount);

			m_free(ALPHA);
			m_free(RETURN_ALPHA);

			mxArray* mx_B = mxCreateNumericMatrix (1,ccount, mxDOUBLE_CLASS, mxREAL);

			memcpy(mxGetPr(mx_B), B, ccount*sizeof(double)) ;

			free_array(B);

			mxArray* mx_SVs = NULL;
			if (!save_memory) {
				mx_SVs = mxCreateNumericMatrix (X_rows, svCount, mxDOUBLE_CLASS, mxREAL);

				for (int i=0; i < svCount; i++) {
					double* pSV = NULL;
					SVs->GetSample(&pSV, i);

					memcpy (mxGetPr(mx_SVs) + i*X_rows, pSV, X_rows * sizeof(double));	
				}
			}
			else mx_SVs = mxCreateNumericMatrix (0, 0, mxDOUBLE_CLASS, mxREAL);

			mxArray* mx_cSV = mxCreateNumericMatrix (1, svCount, mxDOUBLE_CLASS, mxREAL);
			SVs->getClassInfo(mxGetPr(mx_cSV));

			int ucl_length = (ccount > 1) ? ccount : 2;
			mxArray* mx_UCL = mxCreateNumericMatrix (ucl_length, UCL_rows, mxDOUBLE_CLASS, mxREAL);
			m_to_vec (mxGetPr(mx_UCL), UCL, UCL_rows, ucl_length);

			mxArray* mx_nf = mxCreateNumericMatrix (1, 1, mxDOUBLE_CLASS, mxREAL);
			*mxGetPr(mx_nf) = K->NormFactor(); 

			mxArray* mx_kernel_params = mxCreateNumericMatrix (1, K->nParams(), mxDOUBLE_CLASS, mxREAL);
			memcpy(mxGetPr(mx_kernel_params), best + 1, K->nParams()*sizeof(double));

			int* sv_indizes = int_array(svCount);
			SVs->getAbsIndex(sv_indizes);
			for (int i=0; i < svCount;i++) sv_indizes[i]++; // matlab indices are 1 based
			mxArray* mx_sv_indizes = mxCreateNumericMatrix (1,svCount, mxINT32_CLASS, mxREAL); 
			memcpy (mxGetPr(mx_sv_indizes), sv_indizes, svCount*sizeof(int));
			free_array(sv_indizes);

			delete SVs;

			plhs[0] = mx_ALPHA;
			plhs[1] = mx_B;
			plhs[2] = mx_SVs;
			plhs[3] = mx_cSV;
			plhs[4] = mx_UCL;
			plhs[5] = mx_sv_indizes;
			plhs[6] = mx_nf;
			plhs[7] = mx_kernel_params;

			if (return_best_mean_error) {
				mxArray* mx_mean_error = mxCreateNumericMatrix (1, 1, mxDOUBLE_CLASS, mxREAL);
				*mxGetPr(mx_mean_error) = best_mean_error; 
			
				plhs[8] = mx_mean_error;
			}

			if (return_best_mean_sv_count) {
				mxArray* mx_mean_sv_count = mxCreateNumericMatrix (1, 1, mxDOUBLE_CLASS, mxREAL);
				*mxGetPr(mx_mean_sv_count) = best_mean_svcount; 

				plhs[9] = mx_mean_sv_count;
			}

			if (return_C) {
				mxArray* mx_C = mxCreateNumericMatrix (1, 1, mxDOUBLE_CLASS, mxREAL);
				*mxGetPr(mx_C) = best_C; 

				plhs[10] = mx_C;
			}	
		}

		delete cv;

		delete config;
		delete K;
		delete iconfig;
		delete hconfig;

		free_array(c);
		free_array(best);

		delete X;
	}
	else {
		mexPrintf ("Wrong number of input/output parameters (%i/%i). \n", nrhs, nlhs);
		for (int i=0; i < nlhs; i++) plhs[i] = NULL;
	}
}
