#include "utils.h"
#include "cache.h"

#include <cstdlib>
#include <cmath>
#include <cstring>
#include <ctime>

using namespace std;

void check_hash_func_params (int X_columns, int k_max, int C, int** hash_func_params, int *hash_func_params_length, int* delete_hash_func_params) {
	int size = 0;

	if ((*hash_func_params_length != C) && (*hash_func_params_length != 0)) {

		PRINTF ("Ignoring hash func parameters because of wrong size. Size = %i should be %i. \n", *hash_func_params_length, C); 

		*hash_func_params = create_uv_hash_params (X_columns, k_max, C, &size);
		*delete_hash_func_params = 1;

		*hash_func_params_length = C;
	}
	else if (*hash_func_params == NULL) {
		*hash_func_params = create_uv_hash_params (X_columns, k_max, C, &size);
		*delete_hash_func_params = 1;
	}
	else *delete_hash_func_params = 0;
}

int calc_cache_size (int X_columns, int k_max, int C) {

	int cache_size = (int) ceil(sqrt((double)k_max)*(X_columns / C));
	
	return cache_size;
}

int* create_uv_hash_params (int X_columns, int k_max, int C, int* size) {
	
	int cache_size = calc_cache_size(X_columns, k_max, C);

	srand (((unsigned int)time(NULL)) + 1);	

	/* compute random numbers for the universal hash function */
	int* hash_func_params = int_array (C);
	rand_array (hash_func_params, 1, cache_size, C);
	*size = C;

	return hash_func_params;
}

void init_cache (tCache* cache,int size, int max_usage, int entry_size, int key_length, int time_max, int* hash_func_params) {

	/* limit the number of sum cache lines to the maximum number that can be used theoretically */
	cache->size = size;
	cache->max_usage = (max_usage < size) ? max_usage : size;

	cache->entry_size = entry_size;
	cache->key_length = key_length;
	cache->used = 0;

	cache->misses = 0;
	cache->collisions = 0;
	cache->released = 0;

	cache->Values = (double**) pointer_array(cache->size);
	memset (cache->Values, 0, cache->size*sizeof(double*));

	//PRINTF("init_cache: size: %i\n", size);
	cache->Keys = (int**) pointer_array(cache->size);	
	memset (cache->Keys ,0, cache->size*sizeof(int*));

	cache->usage = int_array(cache->size);
	iassign (cache->usage, time_max + 1, cache->size);

	cache->temp = double_array(cache->entry_size);
	cache->key_temp = int_array(key_length);

	cache->time_max = time_max;

	cache->hash_func_params = hash_func_params;
	cache->cache_on = (int) (max_usage > 0);
}

void free_cache (tCache* cache) {
	free_array (cache->temp);
	free_array (cache->key_temp);

	int i=0;
	for (i=0; i < cache->size; i++) {
		if (cache->Values[i] != NULL) {
			free_array (cache->Values[i]);
			free_array (cache->Keys[i]);
		}
	}
	free_array(cache->Values);
	free_array(cache->Keys);
	free_array(cache->usage);

	if (cache->delete_hash_func_params) free_array(cache->hash_func_params);
}


int get_cache_entry (double** x, int** r_key, int time, tCache* cache) {

	if (!cache->cache_on) {
		*x = cache->temp;
		*r_key = cache->key_temp;
		return -1;
	}

	int* key = *r_key;
	int h = 0;

	if (cache->key_length > 1)  // compute the hash value to access the cache
		h = uv_hash_func (cache->hash_func_params, key, cache->key_length, cache->size);
	else h = *key;
	
	

	double* h_entry = cache->Values[h];
	int* h_key = cache->Keys[h];

	if (time >= 0) cache->usage[h] = time; // set the time flag of the cache row accessed

	if ((h_entry != NULL) && (h_key != NULL)) {
		// something was found

		if (memcmp(key, h_key, cache->key_length*sizeof(int)) == 0) {
			// key requested and key found are identical -> cache hit
			*x = h_entry;

			if (*x == NULL) PRINTF ("*x == NULL, File %s, Line %i\n", __FILE__,__LINE__);

			return 1; // values that match the request were found !
		}
		else {
			// key requested and key found are not identical -> cache collision
			cache->collisions++;
			*x     = h_entry;
			*r_key = h_key;

			if (*x == NULL) PRINTF ("*x == NULL, File %s, Line %i\n", __FILE__,__LINE__);
				
			return 0; // values that match the request were not found, but valid values that do not match the request
		}
	}
	else if (time >= 0) {
		// cache row accessed is empty
		
		cache->misses++;
		if (cache->used < cache->max_usage) {
			/* there is still space available to allocate a new cache row */
			double* new_entry = double_array(cache->entry_size);
			int* new_key = int_array (cache->key_length);

			if ((new_entry != NULL) && (new_key != NULL)) {

				cache->used++;	

				cache->Values[h] = new_entry;
				cache->Keys[h] = new_key;

				/* save in *x the values that belong to the requested key */
				*x = new_entry;
			
				/*save the requested key*/
				*r_key = new_key;

				if (*x == NULL) PRINTF ("*x == NULL, File %s, Line %i\n", __FILE__,__LINE__);

				return -1; /* neither values that match the request nor valid values that do not match the request were found*/
			}
			else {
				PRINTF ("Cache allocation failed. Reducing cache size...\n");
				cache->max_usage = cache->used;
				
				if (new_entry != NULL) free_array(new_entry);
				if (new_key != NULL) free_array(new_key);
			}
		}
		
		// we have to release a cache row that was not used for a long period of time
		cache->released++;

		// find the oldest cache row
		int delete_index = imin (cache->usage,cache->size);

		// move the memory to the new location
		cache->Values[h] = cache->Values[delete_index];
		cache->Keys[h] = cache->Keys[delete_index];
	
		// copy the key
		*r_key = cache->Keys[h];

		// delete the old cache entry
		cache->Values[delete_index] = NULL;
		cache->Keys[delete_index] = NULL;
		cache->usage[delete_index] = cache->time_max + 1; //exclude the unused cache row from being choosen for deletion

		// save in *x the values that belong to the requested key
		*x = cache->Values[h];

		
		if (h == delete_index) {// an error occured
			PRINTF ("h == delete_index, File %s, Line %i\n", __FILE__,__LINE__);
			PRINTF ("time: %i, time_max: %i, h: %i\n", time, cache->time_max, h);
		}

		if (*x == NULL) PRINTF ("*x == NULL, File %s, Line %i\n", __FILE__,__LINE__);

		return 0; // values not found, but valid values that do not match the request
	}

	return -1; // time < 0 -> cache in read only mode, desired values could not be found
}
