#include "utils.h"
#include "svm.h"
#include "cross_validation.h"
#include "Data.h"
#include "kernel_func.h"
#include "svm_params.h"

#include <cmath>
#include <cstring>

using namespace std;

CrossValidation::CrossValidation(int k, ClassInfoDataDesc* data, const char** param_names, int nparams, Kernel* kernel,
		SVMconfig* config, HashConfig* hconfig, InfoConfig* iconfig) : data(data), k(k), kernel(kernel), 
										config(config), iconfig(iconfig), hconfig(hconfig), param_names(param_names), nparams(nparams),
										train_subsets(NULL), test_subsets(NULL), balanced_set(NULL) 
{

	train_subsets = (SubsetClassInfoDataDesc**) pointer_array(k);
	test_subsets  = (SubsetClassInfoDataDesc**) pointer_array(k);

	int* ALL = int_array(data->size());
	for (int i=0; i < data->size(); i++) ALL[i] = i;

	int* BALANCED = int_array(data->size());
	memset(BALANCED, 0, data->size()*sizeof(int));

	double* c = double_array(data->size());
	data->getClassInfo(c);

	double* un    = double_array(data->size());	
	int un_length = unique(un, c, data->size());

	int* I = int_array(data->size());

	PRINTF("training data is divided in %i classes.\n", un_length); FLUSH;

	int** S = im_create(k, data->size());

	int* S_length = int_array(k);
	iassign(S_length, 0, k);	

	int max_size=2*data->size();
	for (int i=0; i < un_length; i++) {
		int I_length = find_eq(I, c, un[i], data->size());

		max_size = (I_length/k < max_size) ? I_length/k : max_size;

		
	}

	for (int i=0; i < un_length; i++) {
		int I_length = find_eq(I, c, un[i], data->size());

		PRINTF("class %g has %i members. Selecting subsets of size %i.\n", un[i], I_length, max_size); FLUSH;

		memcpy(BALANCED + i*max_size*k, I, max_size*k*sizeof(int));

		for (int j=0; j < k; j++) {
			memcpy(S[j] + S_length[j], I + j*max_size, max_size*sizeof(int));
			S_length[j] += max_size;
		}
	}

	int* X = int_array(data->size());
	int* Y = int_array(data->size());
	int X_length=0;
	int Y_length=0;

	balanced_set = new SubsetClassInfoDataDesc(BALANCED, un_length*max_size*k, data);	

	for (int i=0; i < k; i++) {
		set_diff(ALL, S[i], data->size(), S_length[i], X, Y, &X_length, &Y_length);
		train_subsets[i] = new SubsetClassInfoDataDesc(X, X_length, data);
		test_subsets[i]  = new SubsetClassInfoDataDesc(S[i], S_length[i], data);
	}

	free_array(X);

	free_array(Y);

	free_array(c);

	free_array(un);

	free_array(I);

	im_free(S);

	free_array(S_length);

	free_array(ALL);
	free_array(BALANCED);
}

CrossValidation::~CrossValidation (void) {

	for (int i=0; i < k; i++) {
		delete train_subsets[i];
		delete test_subsets[i];
	}

	delete balanced_set;

	free_array(train_subsets);
	free_array(test_subsets);
}

void CrossValidation::eval (double* x, double* y) {
	double ret = 0.0;
	double nSVs = 0;

	for (int i=0; i < nparams; i++) {
		kernel->Set(param_names[i], kernel->GridMap(param_names[i],x[i]));
		config->set(param_names[i], config->GridMap(param_names[i],x[i]), train_subsets[i]->size());

		PRINTF("%s : %g ", param_names[i], kernel->GridMap(param_names[i], config->GridMap(param_names[i],x[i]) )); FLUSH;
	}

	PRINTF("| error rates : "); FLUSH;


	for (int i=0; i < k; i++) {

		double* true_labels = double_array(test_subsets[i]->size());
		test_subsets[i]->getClassInfo(true_labels);

		int* errors = int_array(test_subsets[i]->size());

		double** ALPHA=NULL;
		double* B = NULL;
		SubsetClassInfoDataDesc* SVs=NULL;

		double** UCL=NULL;
		int UCL_rows=0;

		int ccount=0;

		int train_ret = svm_wrapper (&ALPHA, &B, &SVs, &UCL, &UCL_rows, &ccount, train_subsets[i], config, kernel, hconfig, iconfig);

		double** DECISION=NULL;
		double* LABELS   =NULL;

		svm_classify (&DECISION, &LABELS, ALPHA, B, SVs, UCL, UCL_rows, ccount, kernel, test_subsets[i], config->do_output());

		sub(LABELS, LABELS, true_labels, test_subsets[i]->size());
		double nerrors = (double) find_neq(errors, LABELS, 0, test_subsets[i]->size());

		double cur_rate = nerrors/((double) test_subsets[i]->size());

		PRINTF("%2.3f ", cur_rate);
		FLUSH;

		ret += cur_rate;
		nSVs += SVs->size();

		m_free(ALPHA);
		free_array(B);
		delete SVs;
		m_free(UCL);

		m_free(DECISION);
		free_array(LABELS);

		free_array(true_labels);
		free_array(errors);
	}

	ret  /= ((double)k);
	nSVs /= ((double)k);

	PRINTF("| mean error: %g | mean #SVs: %g\n", ret, nSVs); FLUSH;

	y[0] = ret;
	y[1] = nSVs;
}

SubsetClassInfoDataDesc* CrossValidation::GetBalancedSet (void) {
	return balanced_set;
}

