/*
grid search algorithm written by Kai Labusch
Copyright (C) 2007 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/

#include "utils.h"
#include <cstdio>
#include <cstring>

#include "grid_search.h"

using namespace std;

GridSearch::GridSearch(double* tmin, double* tmax, double* tcount, double* tstop_tol, double zoom_factor, int ndims, int max_level,
	   TargetFunc* tf) : ndims(ndims), pos(0), f(NULL), X(NULL), max_level(max_level), target_func(tf), zoom_factor(zoom_factor), n_of_evals(0) {

	min      = double_array(ndims);
	max      = double_array(ndims);
	count    = double_array(ndims);
	stop_tol = double_array(ndims);

	memcpy(min, tmin, ndims*sizeof(double));
	memcpy(max, tmax, ndims*sizeof(double));
	memcpy(count, tcount, ndims*sizeof(double));
	memcpy(stop_tol, tstop_tol, ndims*sizeof(double));

	n_of_evals = prod(count, ndims)*max_level;

	//f = double_array(n_of_evals);
	
	f = m_create(n_of_evals, target_func->Dims());
	X = m_create(n_of_evals, ndims);
}

GridSearch::~GridSearch(void) {
	free_array(min);
	free_array(max);
	free_array(count);
	free_array(stop_tol);

	m_free(f);
	m_free(X);
}

double* GridSearch::search(double* best) {
	memset(best, 0, ndims*sizeof(double));

	double* cur_grid_dist = double_array(ndims);
	double* cur_min       = double_array(ndims);
	double* cur_max       = double_array(ndims);
	double* cur_count     = double_array(ndims);

	double* x = double_array(ndims);

	memcpy(cur_min, min, ndims*sizeof(double));
	memcpy(cur_max, max, ndims*sizeof(double));
	memcpy(cur_count, count, ndims*sizeof(double));

	int old_min_pos = -1;
	int min_pos     = -1;
	int level       = 0;

	print_array("stop_tol: ", stop_tol, ndims);

	do {
		print_array("min: ", cur_min, ndims);
		print_array("max: ", cur_max, ndims);
		print_array("count: ", cur_count, ndims);


		// compute current range of grid
		sub  (cur_grid_dist, cur_max, cur_min, ndims);				

		// reduce grid range according to zoom factor
		smultiply(cur_grid_dist, cur_grid_dist, zoom_factor/2.0, ndims);

		old_min_pos = min_pos;

		// evaluate current part of grid
		eval(x, cur_min, cur_max, cur_count, 0);

		for(int i=0; i < ndims; i++) {
			if(cur_grid_dist[i] < stop_tol[i]) {
				cur_count[i]     = 1.0;
				cur_grid_dist[i] = 0.0;
			}
		}

		// find current best parameter combination
		//min_pos = dmin(f, pos);

		double** ft = m_transpose(f, n_of_evals, target_func->Dims());
		int* itemp  = int_array(pos);
		int* itemp2 = int_array(pos);
		
		int* I_cand = int_array(pos);
		for(int i=0;i < pos; i++) I_cand[i]=i;
		int ncand = pos;
	
		double* temp = double_array(pos);

		for(int i=0; i < target_func->Dims();i++) {
			get_i(temp, ft[i], I_cand, ncand);
			min_pos = dmin(temp, ncand);
	
			ncand = find_eq(itemp, temp, temp[min_pos], ncand);
		
			iget_i(itemp2, I_cand, itemp, ncand);
			memcpy(I_cand, itemp2, ncand*sizeof(int));
		}

		min_pos = I_cand[0];

		m_free(ft);
		free_array(itemp);
		free_array(itemp2);
		free_array(I_cand);
		free_array(temp);

		// grid refinement around current best combination
		sub(cur_min, X[min_pos], cur_grid_dist, ndims);
		add(cur_max, X[min_pos], cur_grid_dist, ndims);

		level++;	
	} // stop iteration if (a) maximum number of refinement levels has been reached 
	  //                   (b) target function could not be improved during last refinement
	  //		       (c) stopping tolerance has been reached for all dimensions
	while ((level < max_level) && (min_pos > old_min_pos) && (prod(cur_count, ndims) > 1));

	free_array(cur_grid_dist);
	free_array(cur_min);
	free_array(cur_max);
	free_array(cur_count);
	free_array(x);

	memcpy(best, X[min_pos], ndims*sizeof(double));

	return f[min_pos];
}

void GridSearch::eval(double* x,  double* cur_min, double* cur_max, double* cur_count, int dim) {
	if(dim < ndims) {
		double* cur_dim = double_array(cur_count[dim]);
		linspace(cur_dim, cur_min[dim], cur_max[dim], cur_count[dim]);

	//	print_array("cur_dim: ", cur_dim,  cur_count[dim]);

		for (int i=0; i < cur_count[dim];i++) {
			x[dim] = cur_dim[i];
			eval(x, cur_min, cur_max, cur_count, dim + 1);
		}

		free_array(cur_dim);
	}	
	else {
		target_func->eval(x, f[pos]); // evaluate target function
		memcpy(X[pos], x, ndims*sizeof(double));
		pos++;
	}
}
