/*
(Soft)MaxMinOver learning algorithm written by Kai Labusch
Copyright (C) 2004 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/

#include "kernel_func.h"
#include "utils.h"
#include "svm_params.h"
#include "Data.h"
#include "DiskUtils.h"

#include <cstdio>
#include <cstdlib>
#include <cmath>

using namespace std;

Kernel::Kernel(void) : cur_class(0), P(NULL), precomputation_ID(NULL), pos_temp(NULL), norm_factor(1.0), precomputation_owner(0) {
}

Kernel::Kernel(const Kernel* k) : cur_class(k->cur_class), P(k->P), precomputation_ID(k->precomputation_ID), pos_temp(NULL), norm_factor(k->norm_factor), precomputation_owner(0) {
	if (P != NULL) pos_temp = int_array(precomputation_ID->size());
}

Kernel::~Kernel(void) {
	if ((P != NULL) && precomputation_owner) m_free(P);
	if (pos_temp != NULL) free_array(pos_temp);
}

void Kernel::set_classifier(int i) {
	cur_class = i;
}

int Kernel::get_classifier(void) {
	return cur_class;
}

double Kernel::NormFactor(void) {
	return norm_factor;
}

void Kernel::SetNormFactor(double nf) { 
	norm_factor = nf; 
	PRINTF("Setting kernel normalization factor to: %g\n", norm_factor);
}

void Kernel::load_precomputation(char* filename, DataDesc* pre_ID) {

	precomputation_ID = pre_ID->ID();

	DiskDataDesc* Gdesc = new DiskDataDesc(filename);

	P = m_create(Gdesc->size(), Gdesc->ssize());
	precomputation_owner = 1;

	pos_temp = int_array(Gdesc->ssize());

	int* NZ=NULL;
	
	for(int i=0; i < Gdesc->size(); i++) Gdesc->CopySample(P[i], &NZ, i); 

	delete Gdesc;
}

void Kernel::save_precomputation(char* filename) {

	if ((P != NULL) && (precomputation_ID != NULL )) {
		save_data (filename, P, precomputation_ID->size(), precomputation_ID->size());
	}
}

int Kernel::get_prec (double* dest, DataDesc* a, DataDesc* b, int* pos_a, int pos_a_len, int pos_b, int dest_relative) {

	if ((a->ID() == precomputation_ID) && (b->ID() == precomputation_ID) && (P != NULL)) {
		int* abs_pos_a = pos_temp;
		int abs_pos_b  = 0;

		a->getAbsIndex(abs_pos_a, pos_a, pos_a_len);
		b->getAbsIndex(&abs_pos_b, &pos_b, 1);

		double* pb = P[abs_pos_b]; // P has to be symmetric

		if (!dest_relative) copy_i(dest, pb, pos_a, abs_pos_a, pos_a_len);
		else get_i(dest, pb, abs_pos_a, pos_a_len);

		return 1;
	}

	return 0;
}

Kernel* Kernel::create_kernel(int kernel_type, double* params, int params_length) {

	Kernel* K = NULL;

	switch (kernel_type) {
		case KERNEL_LINEAR : 
			K = new LinK();
		break;
		case KERNEL_RBF : 
			K = new RbfK(params, RbfK::num_of_rbf_params);
		break;
		case KERNEL_L0_RBF : 
			K = new L0_RbfK(params, L0_RbfK::num_of_rbf_params);
		break;

		case KERNEL_POLY :
			if (params != NULL) K = new PolyK(params + 0, params + 1, 1);
			else K = new PolyK(params, params, 1);

		break;	
		default : 
			PRINTF("\nUnknown kernel funktion: %i.\n", kernel_type);
	}

	return K;
}

//-------------------------------------------------------------------------------------------------------------------------------------------

void SCPKernel::precompute(DataDesc* pre_ID) {
	precomputation_ID = pre_ID->ID();

	P        = m_create(pre_ID->size(), pre_ID->size());
	pos_temp = int_array(pre_ID->size());
	precomputation_owner = 1;

	int old_percent=0;

	PRINTF("Performing scp computation ...");FLUSH;

	for (int i=0; i < pre_ID->size(); i++) {

		int new_percent = floor(((double)i*100)/((double) pre_ID->size()));
		if (new_percent != old_percent) {
			old_percent=new_percent;
			if (new_percent % 10 == 0) PRINTF("|");
			else PRINTF(".");
			FLUSH;
		}

		for (int j=i; j < pre_ID->size(); j++) {

			double* xi=NULL;
			double* xj=NULL;

			pre_ID->GetSample(&xi, i);
			pre_ID->GetSample(&xj, j);

			P[i][j] = scp(xi, xj, pre_ID->ssize());
			P[j][i] = P[i][j];

			pre_ID->Flush();
		}
	}
	PRINTF("\n");FLUSH;
}

void SCPKernel::remove_dim_from_precomputation(int d, DimSelectDataDesc* dim_select) {

	DataDesc* source = NULL;
	
	if (dim_select != NULL) {
		if (dim_select->ID() == precomputation_ID) {
			source = dim_select;
			
			int pos=0;
			dim_select->GetDimIndex (&pos, &d, 1);
			
			d=pos;
		}
	}
	else source = precomputation_ID;

	if ((P == NULL) || (precomputation_ID == NULL)) return;

	for (int i=0; i < precomputation_ID->size(); i++) {
		for (int j=i; j < precomputation_ID->size(); j++) {
			double* xi;
			double* xj;

			source->GetSample(&xi, i);
			source->GetSample(&xj, j);	

			P[i][j] -= xi[d]*xj[d];
			P[j][i] = P[i][j];
		
			source->Flush();	
		}
	}
}


void SCPKernel::ComputeNormFactor(DataDesc* data) {

	if (precomputation_ID != NULL) {
		if ((precomputation_ID->ID() == data->ID()) && (P != NULL)) {

			double mn=0.0;

			for(int i=0; i < data->size(); i++) {
				int abs_pos=0;
				data->getAbsIndex(&abs_pos, &i, 1);
		
				mn += sqrt(P[abs_pos][abs_pos]);
			}

			norm_factor = ((double) data->size())/mn;

			return;
		}
	}

	double mn=0.0;
	for (int i=0; i < data->size(); i++) {

		double* xi=NULL;
		data->GetSample(&xi, i);

		mn += sqrt(scp(xi, xi, data->ssize()));

		data->Flush();
	}
	norm_factor = ((double) data->size())/mn;
}

//---------------------------------------------------------------------------------------------------------------------------


void L0_DiffSCPKernel::precompute(DataDesc* pre_ID) {
	precomputation_ID = pre_ID->ID();

	P        = m_create(pre_ID->size(), pre_ID->size());
	pos_temp = int_array(pre_ID->size());
	precomputation_owner = 1;

	int old_percent=0;

	PRINTF("Performing l0_diff_scp precomputation: "); FLUSH;

	for (int i=0; i < pre_ID->size(); i++) {

		int new_percent = floor(((double)i*100)/((double) pre_ID->size()));
		if (new_percent != old_percent) {
			old_percent=new_percent;
			if (new_percent % 10 == 0) PRINTF("|");
			else PRINTF(".");
			FLUSH;
		}

		for (int j=i; j < pre_ID->size(); j++) {

			double* xi=NULL;
			double* xj=NULL;

			pre_ID->GetSample(&xi, i);
			pre_ID->GetSample(&xj, j);

			P[i][j] = l0_diff_scp(xi, xj, pre_ID->ssize());
			P[j][i] = P[i][j];

			pre_ID->Flush();
		}
	}
	PRINTF("\n");
	FLUSH;
}

void L0_DiffSCPKernel::remove_dim_from_precomputation(int d, DimSelectDataDesc* dim_select) {

	DataDesc* source = NULL;
	
	if (dim_select != NULL) {
		if (dim_select->ID() == precomputation_ID) {
			source = dim_select;
			
			int pos=0;
			dim_select->GetDimIndex (&pos, &d, 1);
			
			d=pos;
		}
	}
	else source = precomputation_ID;

	if ((P == NULL) || (precomputation_ID == NULL)) return;

	for (int i=0; i < precomputation_ID->size(); i++) {
		for (int j=i; j < precomputation_ID->size(); j++) {
			double* xi;
			double* xj;

			source->GetSample(&xi, i);
			source->GetSample(&xj, j);

			P[i][j] -= (xi[d] != xj[d]) ? 1 : 0;
			P[j][i] = P[i][j];
		
			source->Flush();	
		}
	}
}

void L0_DiffSCPKernel::ComputeNormFactor(DataDesc* data) {

	norm_factor = 1.0/sqrt(((double) data->ssize()));
	PRINTF("L0_DiffSCPKernel norm factor: %g\n", norm_factor);
	FLUSH;
}


//---------------------------------------------------------------------------------------------------------


void DiffSCPKernel::precompute(DataDesc* pre_ID) {
	precomputation_ID = pre_ID->ID();

	P        = m_create(pre_ID->size(), pre_ID->size());
	pos_temp = int_array(pre_ID->size());
	precomputation_owner = 1;

	int old_percent=0;

	PRINTF("Performing diff_scp precomputation: "); FLUSH;

	for (int i=0; i < pre_ID->size(); i++) {

		int new_percent = floor(((double)i*100)/((double) pre_ID->size()));
		if (new_percent != old_percent) {
			old_percent=new_percent;
			if (new_percent % 10 == 0) PRINTF("|");
			else PRINTF(".");
			FLUSH;
		}

		for (int j=i; j < pre_ID->size(); j++) {

			double* xi=NULL;
			double* xj=NULL;

			pre_ID->GetSample(&xi, i);
			pre_ID->GetSample(&xj, j);

			P[i][j] = diff_scp(xi, xj, pre_ID->ssize());
			P[j][i] = P[i][j];

			pre_ID->Flush();
		}
	}
	PRINTF("\n");
	FLUSH;
}

void DiffSCPKernel::remove_dim_from_precomputation(int d, DimSelectDataDesc* dim_select) {

	DataDesc* source = NULL;
	
	if (dim_select != NULL) {
		if (dim_select->ID() == precomputation_ID) {
			source = dim_select;
			
			int pos=0;
			dim_select->GetDimIndex (&pos, &d, 1);
			
			d=pos;
		}
	}
	else source = precomputation_ID;

	if ((P == NULL) || (precomputation_ID == NULL)) return;

	for (int i=0; i < precomputation_ID->size(); i++) {
		for (int j=i; j < precomputation_ID->size(); j++) {
			double* xi;
			double* xj;

			source->GetSample(&xi, i);
			source->GetSample(&xj, j);

			double diff = xi[d] - xj[d];

			P[i][j] -= diff*diff;
			P[j][i] = P[i][j];
		
			source->Flush();	
		}
	}
}

void DiffSCPKernel::ComputeNormFactor(DataDesc* data) {

	double mn=0.0;
	for (int i=0; i < data->size(); i++) {

		double* xi=NULL;
		data->GetSample(&xi, i);

		mn += sqrt(scp(xi, xi, data->ssize()));

		data->Flush();
	}
	norm_factor = ((double) data->size())/mn;
	PRINTF("DiffSCPKernel norm factor: %g\n", norm_factor);
	FLUSH;
}

//---------------------------------------------------------------------------------------------------------

const int RbfK::num_of_rbf_params = 1;
const char* RbfK::rbf_param_names [] = { "gamma" };

RbfK::RbfK(double* w, int w_len) : DiffSCPKernel(), width(NULL), w_length(w_len) {

	const double Def[] = { 1.0 };
	
	if (!w) {
		w = (double*) Def;
		w_length = 1;
	}

	width = double_array(w_length);
	memcpy(width, w, w_length*sizeof(double));

	cur_width = width[get_classifier()];
}

RbfK::RbfK(const RbfK* k) : DiffSCPKernel(k) {
	width    = double_array(k->w_length);
	w_length = k->w_length;

	memcpy(width, k->width, w_length*sizeof(double));

	cur_width = width[get_classifier()];
}

RbfK::~RbfK(void) {
	free_array(width);
}

void RbfK::Set (const char* name, double val) {
	if (strcmp(name,"gamma") == 0) {
		assign(width, val, w_length);
	}
//	else PRINTF("Unknown parameter: %s (RbfK::Set)\n", name);
}

Kernel* RbfK::Copy (void) {
	return new RbfK(this);
}

void RbfK::set_classifier(int i) {
	if(i < w_length) {
		Kernel::set_classifier(i);
		cur_width = width[get_classifier()];
	}
//	else PRINTF("Could not find kernel parameter for classifier %i\n", i);

//	PRINTF("Current rbf width: %f\n", cur_width);
//	FLUSH;
}

void RbfK::eval(double* dest, DataDesc* X, DataDesc* x, int* I, int I_len, int i) {

	double Gamma = cur_width*NormFactor()*NormFactor();

	if (!get_prec(dest, X, x, I, I_len, i, 0)) {

		double* xi = double_array(x->ssize());
		int* NZ = NULL;

		x->CopySample(xi, &NZ, i);

		for (int j=0;j < I_len; j++) {

			double* xj = NULL;
			X->GetSample(&xj, I[j]);

			dest[I[j]] = exp (-Gamma * diff_scp(xj, xi, X->ssize()));

			X->Flush();
		}
		free_array(xi);
	}
	else {
		for (int j=0;j < I_len; j++) dest[I[j]] = exp (-Gamma * dest[I[j]]);
	}
}

double RbfK::add_rank_info (double** dest, DataDesc* desc_xa, int pos_a, DataDesc* desc_xb, int pos_b, double factor) {

	int X_rows = desc_xa->ssize();
	double Gamma = cur_width*NormFactor()*NormFactor();

	double* xa=NULL;
	double* xb=NULL;
	desc_xa->GetSample(&xa, pos_a);
	desc_xb->GetSample(&xb, pos_b);

	double e=0;

	if (!get_prec(&e, desc_xa, desc_xb, &pos_a, 1, pos_b, 1)) e=diff_scp(xa, xb, X_rows);

	for(int i=0; i < X_rows; i++) *dest[i] += factor*exp(-Gamma*(e - (xa[i] - xb[i])*(xa[i] - xb[i])));

	desc_xa->Flush();
	desc_xb->Flush();

	return factor*exp(-Gamma*e);
}

int RbfK::type(void) { return KERNEL_RBF; }

double RbfK::Param(int i) {
	return width[get_classifier()];
}	

//---------------------------------------------------------------------------------------------------------------------

const int L0_RbfK::num_of_rbf_params = 1;
const char* L0_RbfK::rbf_param_names [] = { "gamma" };

L0_RbfK::L0_RbfK(double* w, int w_len) : L0_DiffSCPKernel(), width(NULL), w_length(w_len) {

	const double Def[] = { 1.0 };
	
	if (!w) {
		w = (double*) Def;
		w_length = 1;
	}

	width = double_array(w_length);
	memcpy(width, w, w_length*sizeof(double));

	cur_width = width[get_classifier()];
}

L0_RbfK::L0_RbfK(const L0_RbfK* k) : L0_DiffSCPKernel(k) {
	width    = double_array(k->w_length);
	w_length = k->w_length;

	memcpy(width, k->width, w_length*sizeof(double));

	cur_width = width[get_classifier()];
}

L0_RbfK::~L0_RbfK(void) {
	free_array(width);
}

void L0_RbfK::Set (const char* name, double val) {
	if (strcmp(name,"gamma") == 0) {
		assign(width, val, w_length);
	}
//	else PRINTF("Unknown parameter: %s (RbfK::Set)\n", name);
}

Kernel* L0_RbfK::Copy (void) {
	return new L0_RbfK(this);
}

void L0_RbfK::set_classifier(int i) {
	if(i < w_length) {
		Kernel::set_classifier(i);
		cur_width = width[get_classifier()];
	}
//	else PRINTF("Could not find kernel parameter for classifier %i\n", i);

//	PRINTF("Current rbf width: %f\n", cur_width);
//	FLUSH;
}

void L0_RbfK::eval(double* dest, DataDesc* X, DataDesc* x, int* I, int I_len, int i) {

	double Gamma = cur_width*NormFactor()*NormFactor();

	if (!get_prec(dest, X, x, I, I_len, i, 0)) {

		double* xi   = double_array(x->ssize());
		int* NZ=NULL;

		x->CopySample(xi, &NZ, i);

		for (int j=0;j < I_len; j++) {

			double* xj = NULL;		
			X->GetSample(&xj, I[j]);

			dest[I[j]] = exp (-Gamma * l0_diff_scp(xj, xi, X->ssize()));

			X->Flush();
		}
		free_array(xi);
	}
	else {
		for (int j=0;j < I_len; j++) dest[I[j]] = exp (-Gamma * dest[I[j]]);
	}
}

double L0_RbfK::add_rank_info (double** dest, DataDesc* desc_xa, int pos_a, DataDesc* desc_xb, int pos_b, double factor) {

	int X_rows = desc_xa->ssize();
	double Gamma = cur_width*NormFactor()*NormFactor();

	double* xa=NULL;
	double* xb=NULL;
	desc_xa->GetSample(&xa, pos_a);
	desc_xb->GetSample(&xb, pos_b);

	double e=0;
	double ret=0;

	if (!get_prec(&e, desc_xa, desc_xb, &pos_a, 1, pos_b, 1)) e=l0_diff_scp(xa, xb, X_rows);

	if (factor > 0) {
		factor = log(factor);

		for(int i=0; i < X_rows; i++) *dest[i] += exp(-Gamma*(e - ((xa[i] != xb[i]) ? 1.0 : 0)) + factor);

		ret = exp(-Gamma*e + factor);
	}
	else if (factor < 0) {
		factor = log(-factor);

		for(int i=0; i < X_rows; i++) *dest[i] += -exp(-Gamma*(e - ((xa[i] != xb[i]) ? 1.0 : 0)) + factor);

		ret = -exp(-Gamma*e + factor);
	}

	desc_xa->Flush();
	desc_xb->Flush();

	return ret;
}

int L0_RbfK::type(void) { return KERNEL_L0_RBF; }

double L0_RbfK::Param(int i) {
	return width[get_classifier()];
}	

//---------------------------------------------------------------------------------------------------------------------


const int LinK::num_of_link_params=0;
const char* LinK::link_param_names[]={};

LinK::LinK(void) : SCPKernel() {
}

LinK::LinK(const LinK* k) : SCPKernel(k) { 
}

LinK::~LinK(void){
}

void LinK::Set (const char* name, double val) {
}

Kernel* LinK::Copy (void) {
	return new LinK(this);
}

void LinK::eval(double* dest, DataDesc* X, DataDesc* x, int* I, int I_len, int i) {

	if (!get_prec(dest, X, x, I, I_len, i, 0)) {

		double* xi = double_array(x->ssize());
		int* NZ=NULL;

		x->CopySample(xi, &NZ, i);

		double f = pow(NormFactor(),2);

		for (int j=0; j < I_len; j++) {
			double* xj = NULL;
			X->GetSample(&xj, I[j]);
			dest[I[j]] = scp(xi, xj, X->ssize())*f;

			X->Flush();
		}

		free_array(xi);
	}
	else smultiply_i(dest, dest, pow(NormFactor(),2), I, I_len);
}

double LinK::add_rank_info (double** dest, DataDesc* desc_xa, int pos_a, DataDesc* desc_xb, int pos_b, double factor) {

	int X_rows = desc_xa->ssize();

	double* xa=NULL;
	double* xb=NULL;
	desc_xa->GetSample(&xa, pos_a);
	desc_xb->GetSample(&xb, pos_b);

	double e=0;

	factor = factor*pow(NormFactor(),2);

	if (!get_prec(&e, desc_xa, desc_xb, &pos_a, 1, pos_b, 1)) e=scp(xa, xb, X_rows);
	
	double* limit = xa + X_rows;

	while (xa < limit - 4) {
		**dest++ += factor*(e - *xa++ * *xb++);
		**dest++ += factor*(e - *xa++ * *xb++);
		**dest++ += factor*(e - *xa++ * *xb++);
		**dest++ += factor*(e - *xa++ * *xb++);
	}

	while (xa < limit) **dest++ += factor*(e - *xa++ * *xb++);

	desc_xa->Flush();
	desc_xb->Flush();

	return factor*e;
}

int LinK::type(void) { return KERNEL_LINEAR; }

double LinK::Param(int i) {
	return 0;
}

const int PolyK::num_of_polyk_params=2;
const char* PolyK::polyk_param_names [] = { "degree", "k" };

PolyK::PolyK(double* d, double* ck, int len) : SCPKernel(), length(len), degree(NULL), k(NULL) {

	const double DefDegree [] = { 2.0 };
	const double DefK [] = {0.0};

	if ((!d) || (!ck)) {
		d = (double*) DefDegree;
		ck = (double*) DefK;
		length = 1;
	}

	degree = double_array(length);
	memcpy(degree, d, length*sizeof(double));
	k = double_array(length);
	memcpy(k, ck, length*sizeof(double));

	cur_params[0] = degree[get_classifier()];
	cur_params[1] = k[get_classifier()];
}

PolyK::PolyK(const PolyK* kernel) : SCPKernel(kernel), degree(NULL), k(NULL) {
	length = kernel->length;
	degree = double_array(length);
	k = double_array(length);
	
	memcpy(degree, kernel->degree, length*sizeof(double));
	memcpy(k, kernel->k, length*sizeof(double));

	cur_params[0] = degree[get_classifier()];
	cur_params[1] = k[get_classifier()];
}

PolyK::~PolyK(void) {
	free_array(degree);
	free_array(k);
}

void PolyK::Set (const char* name, double val) {
	if (strcmp(name,"d") == 0) {
		assign(degree, val, length);
	}
	else if (strcmp(name,"k") == 0) {
		assign(k, val, length);
	}
	else PRINTF("Unknown parameter: %s (PolyK::Set)\n", name);
}

Kernel* PolyK::Copy (void) {
	return new PolyK(this);
}


void PolyK::set_classifier(int i) {
	if(i < length) {
		Kernel::set_classifier(i);
		cur_params[0] = degree[get_classifier()];
		cur_params[1] = k[get_classifier()];
	}
	else PRINTF("Could not find kernel parameter for classifier %i\n", i);
}

void PolyK::eval(double* dest, DataDesc* X, DataDesc* x, int* I, int I_len, int i) {

	double d = cur_params[0];
	double c = cur_params[1];

	if (!get_prec(dest, X, x, I, I_len, i, 0)) {

		double* xi = double_array(x->ssize());
		int* NZ=NULL;
		x->CopySample(xi, &NZ, i);

		for (int j=0; j < I_len; j++) {
			double* xj = NULL;
			X->GetSample(&xj, I[j]);
			dest[I[j]] = scp(xi, xj, X->ssize())*NormFactor();

			X->Flush();
		}
		free_array(xi);
	}

	double f = pow(NormFactor(), 2*d);

	for (int j=0; j < I_len; j++) dest[I[j]] = pow(f*(dest[I[j]] + c), d);
}

double PolyK::add_rank_info (double** dest, DataDesc* desc_xa, int pos_a, DataDesc* desc_xb, int pos_b, double factor) {

	int X_rows = desc_xa->ssize();

	double* xa=NULL;
	double* xb=NULL;

	desc_xa->GetSample(&xa, pos_a);
	desc_xb->GetSample(&xb, pos_b);

	double d = degree[get_classifier()];
	double c = k[get_classifier()];

	double e=0;
	if (!get_prec(&e, desc_xa, desc_xb, &pos_a, 1, pos_b, 1)) e=scp(xa, xb, X_rows);

	e+=c;

//	int i=0;
//	for (i=0; i < X_rows; i++) dest[i] += factor*pow(e - xa[i]*xb[i] + c, d);

	double* limit = xa + X_rows;
	double f = pow(NormFactor(), 2*d);

	while (xa < limit) **dest++ += factor*pow(f*(e - *xa++ * *xb++), d);

	desc_xa->Flush();
	desc_xb->Flush();

	return factor*pow(f*e, d);
}

int PolyK::type(void) { return KERNEL_POLY; }

double PolyK::Param(int i) {

	if(i == 0) return degree[get_classifier()];
	else if(i== 1) return k[get_classifier()];
	else return 0;
}
