/*
(Soft)MaxMinOver learning algorithm written by Kai Labusch
Copyright (C) 2004 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/

#ifndef __KERNEL_FUNC_H_INCLUDED__
#define __KERNEL_FUNC_H_INCLUDED__

#define KERNEL_LINEAR 0
#define KERNEL_RBF    1
#define KERNEL_POLY   2
#define KERNEL_L0_RBF 3

#include <cstdlib>
#include <cstring>
#include <cmath>

using namespace std;

class DataDesc;
class DimSelectDataDesc;

class Kernel {
public:
	static Kernel* create_kernel(int kernel_type, double* params=NULL, int params_length=0);

	Kernel(void);

	Kernel(const Kernel* k);

	virtual ~Kernel(void);

	// evaluates the kernel with respect to the members of X specified by I and the member of x specified by i
	virtual void eval(double* dest, DataDesc* X, DataDesc* x, int* I, int I_len, int i) = 0;

	// removes each data dimension one-by-one and evaluates the kernel on the reduced data, stores the result of all evaluations
	virtual double add_rank_info (double** dest, DataDesc* xa, int pos_a, DataDesc* xb, int pos_b, double factor) = 0; 

	virtual void set_classifier(int i);
	int get_classifier(void);

	// returns the type identifier of the kernel
	virtual int type(void) = 0;

	// return the number of hyper-parameters of the kernel	
	virtual int nParams(void) = 0;	
	// returns the current value of the i-th hyper-parameter
	virtual double Param(int i) = 0;
	// set the hyper-parameter that is identified by "name" to a certain value
	virtual void Set (const char* name, double val) = 0;
	// returns the names of all hyper-parameters of the kernel
	virtual const char** ParamNames(void) = 0;

	// This function is used during grid search. It maps a grid-value to the final value that is actually used (implements a logrithmic grid for example)
	virtual double GridMap(const char* name, double val) = 0;
	// Provides minimum of search range for the parameter identified by "name" during grid search
	virtual double GridMin(const char* name, double val) = 0;
	// Provides maximum of search range for the parameter identified by "name" during grid search 
	virtual double GridMax(const char* name, double val) = 0;
	// Provides the stopping tolerance for the parameter identified by "name" during grid search (count is the number of grid points in that search dimension)
	virtual double GridStopTol (const char* name, double val, int count) = 0;

	// Provides a deep Copy of the Kernel
	virtual Kernel* Copy (void) = 0;

	// precomputes the part of the kernel matrix that is independent from the hyperparameters with respect to a certain data-set
	// It is important that this function is called with the primal data descriptor where all other subset data descriptors point to
	virtual void precompute(DataDesc* pre_ID) = 0;
	
	virtual void load_precomputation(char* filename, DataDesc* pre_ID);
	virtual void save_precomputation(char* filename);

	virtual int get_prec (double* sc, DataDesc* a, DataDesc* b, int* pos_a, int pos_a_len, int pos_b, int dest_relative);

	// removes the influence of dimension d from the kernel matrix, via dim_select one can provide an alternative source
	// for the required data in order to prevent disk-access if the computation runs on high-dimensional data
	virtual void remove_dim_from_precomputation(int d, DimSelectDataDesc* dim_select=NULL) = 0;

	virtual void ComputeNormFactor(DataDesc* data) = 0;
	double NormFactor(void);
	void SetNormFactor(double nf);
protected:
	double** P;
	DataDesc* precomputation_ID;
	int* pos_temp;
	int precomputation_owner;

	double norm_factor;
private:
	int cur_class;	
};

class SCPKernel : public Kernel {
public:
	SCPKernel(void) : Kernel() { }

	SCPKernel(const SCPKernel* k) : Kernel(k) {}

	virtual void precompute(DataDesc* pre_ID);

	virtual void remove_dim_from_precomputation(int d, DimSelectDataDesc* dim_select=NULL);

	virtual void ComputeNormFactor(DataDesc* data);
};

class DiffSCPKernel : public Kernel {
public:
	DiffSCPKernel(void) : Kernel() { }

	DiffSCPKernel(const DiffSCPKernel* k) : Kernel(k) { }

	virtual void precompute(DataDesc* pre_ID);

	virtual void remove_dim_from_precomputation(int d, DimSelectDataDesc* dim_select=NULL);

	virtual void ComputeNormFactor(DataDesc* data);
};


class L0_DiffSCPKernel : public Kernel {
public:
	L0_DiffSCPKernel(void) : Kernel() { }

	L0_DiffSCPKernel(const L0_DiffSCPKernel* k) : Kernel(k) { }

	virtual void precompute(DataDesc* pre_ID);

	virtual void remove_dim_from_precomputation(int d, DimSelectDataDesc* dim_select=NULL);

	virtual void ComputeNormFactor(DataDesc* data);
};


class RbfK : public DiffSCPKernel {
public:
	RbfK(double* w, int w_len);
	RbfK(const RbfK* k);
	virtual ~RbfK(void);

	virtual void set_classifier(int i);

	// evaluates the kernel with respect to the members of X specified by I and the member of x specified by i
	virtual void eval(double* dest, DataDesc* X, DataDesc* x, int* I, int I_len, int i);

	virtual double add_rank_info (double** dest, DataDesc* xa, int pos_a, DataDesc* xb, int pos_b, double factor);
	virtual int type(void);

	virtual double Param(int i);
	virtual void Set (const char* name, double val);
	virtual Kernel* Copy (void);

	virtual int nParams(void) { return num_of_rbf_params; }	
	virtual const char** ParamNames (void) { return (const char**) rbf_param_names; }

	virtual double GridMap(const char* name, double val) {
		if (strcmp(name, "gamma")==0) return pow(10,val);
		return val;
	}

	virtual double GridMin(const char* name, double val) {
		if(strcmp(name,"gamma")==0) return -2.0;		
		return val;
	}

	virtual double GridMax(const char* name, double val) {
		if (strcmp(name,"gamma")==0) return 2.0;
		return val;
	}

	virtual double GridStopTol (const char* name, double val, int count) {
		if (strcmp(name,"gamma")==0) return 0.1*((double)count);
		return val;
	}

	static const int num_of_rbf_params;
	static const char* rbf_param_names [];
private:
	double* width;
	int w_length;

	double cur_width;
}; 

class L0_RbfK : public L0_DiffSCPKernel {
public:
	L0_RbfK(double* w, int w_len);
	L0_RbfK(const L0_RbfK* k);
	virtual ~L0_RbfK(void);

	virtual void set_classifier(int i);

	// evaluates the kernel with respect to the members of X specified by I and the member of x specified by i
	virtual void eval(double* dest, DataDesc* X, DataDesc* x, int* I, int I_len, int i);

	virtual double add_rank_info (double** dest, DataDesc* xa, int pos_a, DataDesc* xb, int pos_b, double factor);
	virtual int type(void);

	virtual double Param(int i);
	virtual void Set (const char* name, double val);
	virtual Kernel* Copy (void);

	virtual int nParams(void) { return num_of_rbf_params; }	
	virtual const char** ParamNames (void) { return (const char**) rbf_param_names; }

	virtual double GridMap(const char* name, double val) {
		if (strcmp(name, "gamma")==0) return pow(10,val);
		return val;
	}

	virtual double GridMin(const char* name, double val) {
		if(strcmp(name,"gamma")==0) return -2.0;		
		return val;
	}

	virtual double GridMax(const char* name, double val) {
		if (strcmp(name,"gamma")==0) return 2.0;
		return val;
	}

	virtual double GridStopTol (const char* name, double val, int count) {
		if (strcmp(name,"gamma")==0) return 0.1*((double)count);
		return val;
	}

	static const int num_of_rbf_params;
	static const char* rbf_param_names [];
private:
	double* width;
	int w_length;

	double cur_width;
}; 


class LinK : public SCPKernel {
public:
	LinK(void);
	LinK(const LinK* k);
	virtual ~LinK(void);

	// evaluates the kernel with respect to the members of X specified by I and the member of x specified by i
	virtual void eval(double* dest, DataDesc* X, DataDesc* x, int* I, int I_len, int i);

	virtual double add_rank_info (double** dest, DataDesc* xa, int pos_a, DataDesc* xb, int pos_b, double factor);
	virtual int type(void);

	virtual double Param(int i);
	virtual void Set (const char* name, double val);
	virtual Kernel* Copy (void);

	virtual int nParams(void) { return num_of_link_params; }	
	virtual const char** ParamNames (void) { return (const char**) link_param_names; }

	virtual double GridMap(const char* name, double val) {
		return val;
	}

	virtual double GridMin(const char* name, double val) {
		return val;
	}

	virtual double GridMax(const char* name, double val) {
		return val;
	}

	virtual double GridStopTol (const char* name, double val, int count) {
		return val;
	}

private:
	static const int num_of_link_params;
	static const char* link_param_names[];
};

class PolyK : public SCPKernel {
public:
	PolyK(double* d, double* ck, int len);
	PolyK(const PolyK* k);
	virtual ~PolyK(void);
	virtual void set_classifier(int i);

	// evaluates the kernel with respect to the members of X specified by I and the member of x specified by i
	virtual void eval(double* dest, DataDesc* X, DataDesc* x, int* I, int I_len, int i);

	virtual double add_rank_info (double** dest, DataDesc* xa, int pos_a, DataDesc* xb, int pos_b, double factor);
	virtual int type(void);

	virtual double Param(int i);			
	virtual void Set (const char* name, double val);
	virtual Kernel* Copy (void);

	virtual int nParams(void) { return num_of_polyk_params; }	
	virtual const char** ParamNames (void) { return (const char**) polyk_param_names; }

	virtual double GridMap(const char* name, double val) {
		if (strcmp(name,"degree")==0) return floor(val);
		return val;
	}

	virtual double GridMin(const char* name, double val) {
		if (strcmp(name,"degree")==0) return 2.0;
		if (strcmp(name,"k")==0) return 0.0;
		return val;
	}

	virtual double GridMax(const char* name, double val) {
		if (strcmp(name,"degree")==0) return 10.0;
		if (strcmp(name,"k")==0) return 1.0;
		return val;
	}

	virtual double GridStopTol (const char* name, double val, int count) {
		if (strcmp(name,"degree")==0) return count;
		if (strcmp(name,"k")==0) return 0.1*((double)count);
		return val;
	}


private:
	static const int num_of_polyk_params;
	static const char* polyk_param_names[];

	double* degree;
	double* k;
	int length;

	double cur_params[2];
};

#endif
