/*
(Soft)DoubleMaxMinOver learning algorithm written by Kai Labusch
Copyright (C) 2004 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software


Contact:
labusch@inb.uni-luebeck.de
*/

#include "utils.h"
#include "cache.h"
#include "min.h"
#include "svm.h"
#include "kernel_func.h"
#include "svm_params.h"
#include "svm_utils.h"
#include "DiskUtils.h"
#include "Data.h"

#include <cmath>
#include <cstring>
#include <ctime>
#include <cstdlib>
#include <unistd.h>
#include <cstdio>

using namespace std;

#ifdef IN
#undef IN
#endif 

void svm_classify (double*** r_DECISION, double** r_labels, double** ALPHA, double* B, 
		   DataDesc* SVs, double** UCL, int UCL_rows, int ccount, Kernel* K, DataDesc* X, int output) {

	int svCount   = SVs->size();
	int X_rows    = X->ssize();
	int X_columns = X->size(); 

	double* temp = double_array (svCount);
	memset(temp, 0, svCount*sizeof(double)); 

	double* temp2 = double_array (ccount);
	int* itemp2   = int_array(svCount);	

	int calc_len = 0;
	int* calc    = int_array(svCount);

	double** DECISION = NULL;

	if (*r_DECISION == NULL) {
		DECISION    = m_create(X_columns, ccount);
		*r_DECISION = DECISION;
	}
	else DECISION = *r_DECISION;

	m_zero(DECISION, X_columns, ccount);

	double* decision = NULL;

	for(int j=0; j < ccount; j++) { // do this for every classifier

		K->set_classifier(j); // notify the kernel about the current classifier

		// determine the SVs that are currently active, i.e. that have non-zero alphas
		calc_len = find_neq(calc, ALPHA[j], 0, svCount);

		for(int i=0; i < X_columns; i++) { // do the following for every data sample to be classified
			decision = DECISION[i]; // decision contains the decision values of all classifiers for the current test sample

			// compute the kernel values for the data to be classified
			K->eval(temp, SVs, X, calc, calc_len, i);	

			// multiply kernel values by their weights , compute decision value
			multiply (temp, temp, ALPHA[j], svCount); // do not set temp to zero (ALPHA contains the zeros) !!!

			decision[j] = sum(temp, svCount); // sum over the contributions of each SV
		}

		SVs->Flush();
	}

	double* labels = double_array (X_columns);

	if (ccount > 1) { // multi class case

		if (output) { 
			PRINTF("multi class case: "); FLUSH;
		}

		int strategy = (UCL_rows > 1) ? ONE_AGAINST_ONE : ONE_AGAINST_ALL;

		if (strategy == ONE_AGAINST_ALL) {
	
			if(output) PRINTF ("ONE_AGAINST_ALL\n"); FLUSH;

			for(int i=0; i < X_columns; i++) {
				decision = DECISION[i];
				// add threshold
				add(temp2, decision, B, ccount); 

				// compute class labels take the class having the maxmimum decision value
				labels[i] = UCL[0][dmax(temp2, ccount)]; 
			}	
		}
		else if (strategy == ONE_AGAINST_ONE) {

			if(output) PRINTF ("ONE_AGAINST_ONE.\n"); FLUSH;

			double* un = double_array(ccount* UCL_rows);
			double* UCL_vec = double_array(UCL_rows * ccount);
			m_to_vec(UCL_vec,UCL, UCL_rows, ccount);
			int un_length = unique(un, UCL_vec, ccount*UCL_rows);
			free_array(UCL_vec);

			double* counter = double_array(un_length);
			double** MAP = m_create(UCL_rows, ccount);
			for (int k=0; k < ccount; k++) { // create map from class labels to counter indices
				int j=0;	
				while(un[j] != UCL[0][k]) j++;
				MAP[0][k] = (double)j;	

				j=0;	
				while(un[j] != UCL[1][k]) j++;
				MAP[1][k] = (double)j;
			}

			for(int i=0; i < X_columns; i++) {
				decision = DECISION[i];
				// add threshold
				add(temp2, decision, B, ccount);
			
				memset(counter, 0, un_length*sizeof(double));
		
				// determine the decision of every classifier
				for (int k=0; k < ccount; k++) counter[(int) ((temp2[k] >= 0) ? MAP[0][k] : MAP[1][k])]++;

				// determine the final decision i.e. the class that was choosen by the majority of the classifiers
				labels[i] = un[dmax(counter,un_length)];
			}
			m_free(MAP);
			free_array(un);
			free_array(counter);
		}
		else if (output) PRINTF("unknown multi class strategy: %i\n", strategy);
	}
	else { // two-class case

		if (output) PRINTF("two class case.\n");
		
		// compute class labels
		for(int i=0; i < X_columns; i++) labels[i] = (*(DECISION[i]) + B[0] >= 0) ? UCL[0][0] : UCL[0][1];
	}

	SVs->Flush();
	free_array (temp);
	free_array (temp2);
	free_array (itemp2);

	free_array (calc);

	if (r_labels != NULL) {
		if (*r_labels == NULL) *r_labels = labels;
		else {
			memcpy(*r_labels, labels, X_columns*sizeof(double));
			free_array(labels);
		}
	}
	else free_array(labels);
}

// rank features according to their importance for the margin
void rank_features_margin (double*** r_RANK, double** ALPHA, double* B, DataDesc* SVs, int ccount, Kernel* K, int output) {

	int svCount   = SVs->size(); 
	int X_rows    = SVs->ssize();

	double* temp = double_array (svCount);
	memset(temp, 0, svCount*sizeof(double)); 

	double* temp2 = double_array (X_rows);

	int* calc    = int_array(svCount);
	int calc_len = 0;

	double** RANK = NULL;
	double* new_margin = double_array(X_rows);

	if (*r_RANK== NULL) {
		RANK = m_create(ccount, X_rows);
		*r_RANK = RANK;
	}
	else RANK = *r_RANK;

	m_zero(RANK, ccount, X_rows);

	for(int j=0; j < ccount; j++) { // every classifier

		if (output) PRINTF("classifier: %i\n", j);

		double* alpha = ALPHA[j];

		K->set_classifier(j);
		
		// determine the SVs that are currently active
		calc_len = find_neq(calc, alpha, 0, svCount);

		double old_margin = 0.0;

		double** TEMP = m_create(X_rows, calc_len);
		double** DEST = (double**) pointer_array(X_rows);

		double** NEW_MARGIN = m_create(X_rows, calc_len);
		m_zero(NEW_MARGIN, X_rows, calc_len);

		memset(new_margin, 0, X_rows*sizeof(double));
			
		double old_percent=0.0;
		for(int i=0; i < calc_len; i++) { // for every sv to be classified 

			double percent = floor( ((double)i)*100/((double)calc_len));
			if (old_percent != percent) {
				old_percent = percent;
				PRINTF("%g%% ", percent);
				FLUSH;

				if (((int) percent) % 10 == 0) {
					PRINTF("\n");FLUSH;
				}
			}

			// compute the kernel values for the sv to be classified	
			K->eval(temp, SVs, SVs, calc, calc_len, calc[i]);	

			// multiply kernel values by their weights 
			multiply (temp, temp, alpha, svCount); // do not set temp to zero -> alpha contains appropriate zeros!!!
			isort(temp, NULL, svCount);

			old_margin += alpha[calc[i]]*sum(temp, svCount); // compute norm of normal vector based on all features
			
			memset(temp2, 0, X_rows*sizeof(double));

			m_zero(TEMP, X_rows, calc_len);
			memcpy(DEST, TEMP, X_rows*sizeof(double**));

			// now evaluate the impact of each single feature by removing it from the normal vector 
			//K->add_rank_info(temp2, SVs, calc[i], SVs, calc[i], alpha[calc[i]]*alpha[calc[i]]); // calculate leave one out kernel values
			
			K->add_rank_info(DEST, SVs, calc[i], SVs, calc[i], alpha[calc[i]]*alpha[calc[i]]); // calculate leave one out kernel values
			psadd(DEST, DEST, 1, X_rows);

			for (int k=i+1; k < calc_len; k++) { // do it for every support vector
				//K->add_rank_info(temp2, SVs, calc[k], SVs, calc[i], 2*alpha[calc[k]]*alpha[calc[i]]); // calculate leave one out kernel values
	
				K->add_rank_info(DEST, SVs, calc[k], SVs, calc[i], 2*alpha[calc[k]]*alpha[calc[i]]); // calculate leave one out kernel values

				psadd(DEST, DEST, 1, X_rows);
			}

			// new_margin contains the norm of the normal vector if a certain dimension is removed 
			//add(new_margin, new_margin, temp2, X_rows); // the norm of the normal vector
			//				            // is the sum of all kernel values

			for (int k=0; k < X_rows; k++) {
				isort(TEMP[k], NULL, calc_len - i);
				//new_margin[k] += sum(TEMP[k], calc_len - i);

				NEW_MARGIN[k][i] = sum(TEMP[k], calc_len - i);
			}
		}
		PRINTF("\n");FLUSH;

		for (int k=0;k < X_rows;k++) {
			isort(NEW_MARGIN[k], NULL, calc_len);
			new_margin[k] = sum(NEW_MARGIN[k], calc_len);
		}

		// determine the deviation of the norm of the reduced normal vector from the norm of the normal vector
		// that is based on all features , collect the result in RANK (j describes the current classifier)
		smultiply_ip(new_margin, -1.0, X_rows);
		sadd(RANK[j], new_margin, old_margin, X_rows);

		SVs->Flush();

		m_free(TEMP);
		free_array(DEST);
		m_free(NEW_MARGIN);
	}

	free_array (temp);
	free_array (temp2);
	free_array (calc);
	free_array (new_margin);
}

// rank features according to their importance for the classification of the samples in X
void rank_features (double*** r_RANK, double** ALPHA, double* B, DataDesc* SVs, double** UCL, int UCL_rows, int ccount,
		   Kernel* K, ClassInfoDataDesc* X) {

	int X_rows    = X->ssize();
	int X_columns = X->size();

	int svCount = SVs->size();

	double* c = double_array(X_columns);
	X->getClassInfo(c);

	double* temp = double_array (svCount);
	memset(temp, 0, svCount*sizeof(double)); 

	double* temp3	    = double_array (X_rows);

	int* calc           = int_array(svCount);

	double** RANK = NULL;
	double* rank_decision = double_array(X_rows);
	double** DEST = (double**) pointer_array(X_rows);
	
	for (int i=0; i < X_rows; i++) DEST[i] = rank_decision + i;	

	if (*r_RANK== NULL) {
		RANK = m_create(ccount, X_rows);
		*r_RANK = RANK;
	}
	else RANK = *r_RANK;

	m_zero(RANK, ccount, X_rows);

	double** CL = class_assign(c, X_columns, UCL, UCL_rows, ccount);

	for(int j=0; j < ccount; j++) { // every classifier

		K->set_classifier(j);

		// determine the SVs that are currently active
		int calc_len = find_neq(calc, ALPHA[j], 0, svCount);

		for(int i=0; i < X_columns; i++) { // every data sample to be classified

			if (CL[j][i] != 0) { // if the current sample belongs to the current classifier
				// compute the kernel values for the sample to be classified
				K->eval(temp, SVs, X, calc, calc_len, i);	

				// multiply kernel values by their weights
				multiply (temp, temp, ALPHA[j], svCount); // do not set temp to zero !!! ALPHA already contains the zeros!

				double decision = sum(temp, svCount); // decision compute decision value based on all features
		
				// now evaluate the impact of each single feature by removing it from the decision function
				memset(rank_decision, 0, X_rows*sizeof(double));

				for (int k=0; k < calc_len; k++) { // do it for every support vector
					K->add_rank_info(DEST, SVs, calc[k], X, i, ALPHA[j][calc[k]]); // calculate leave one out kernel values
				}

				// determine the deviation of the reduced decision from the decision that is based on all features
				ssub(rank_decision, rank_decision, decision, X_rows); 
				absol(rank_decision, rank_decision, X_rows);
	
				// collect the result in RANK (j describes the current classifier)
				add(RANK[j], RANK[j], rank_decision, X_rows);		
			}
		}
		SVs->Flush();
		X->Flush();
	}

	m_free(CL);

	free_array (temp);
	free_array (temp3);
	free_array (calc);
	free_array (rank_decision);
	free_array(DEST);
	free_array (c);
}

// perform -1/1 assignment that corresponds to chosen multi-class classifier mode
double** class_assign(double* c, int c_length, double** UN, int UN_rows, int ccount) {

	int i=0;
	int j=0;
	int n=0;

	int* itemp = int_array(c_length);

	double** CL = NULL;

	if (UN_rows == 1) { // one against all

		// allocate matrix to contain the two class information for each classifier to be trained
		CL = m_create(ccount, c_length);
		m_assign(CL, ccount, c_length, -1.0);

		// determine the two class information for each classfier to be trained
		for (i=0; i < ccount; i++) {

			n = find_eq(itemp, c, UN[0][i], c_length);

			if (n <= 0) {
				PRINTF ("No data available for class: %f\n", UN[0][i]);
				break;
			}

			assign_i(CL[i], itemp, 1, n); 
		}
	}
	else if (UN_rows == 2) { // one against one

		CL = m_create(ccount, c_length);
		m_zero(CL, ccount, c_length);

		int pos = 0;
		for(pos = 0; pos < ccount; pos++) {
			n = find_eq(itemp, c, UN[0][pos], c_length);		
			assign_i(CL[pos],itemp, 1, n);

			n = find_eq(itemp, c, UN[1][pos], c_length);		
			assign_i(CL[pos],itemp, -1, n);
		}
	}
	else PRINTF("wrong UN information...\n");

	free_array(itemp);

	return CL;
}

// check if classifier configuration is sound with respect to given training data, provide -1/1 two class assignment that corresponds
// to chosen classifier configuration
// return the number of two-class classifiers
int check_class_information (double*** r_CL, double*** r_UN, int* r_UN_rows,  double* c, int c_length, SVMconfig* config) {

	int ret = 1;
	int i=0;
	int j=0;
	
	double** UN = NULL;
	double* un = double_array(c_length);
	int* itemp = int_array(c_length); 

	// determine the number of classes
	int un_length = unique(un, c, c_length);

	// if there are only two classes present, we only need 1 classifier
	int ccount = (un_length > 2) ? un_length : 1;

	int strategy = ONE_AGAINST_ONE;
	double** CL = NULL;

	strategy = (ccount > 1) ? config->mc_strategy() : ONE_AGAINST_ALL;

	if (strategy == ONE_AGAINST_ALL) {

		*r_UN_rows = 1;
		UN = m_create(*r_UN_rows, un_length);
		memcpy(UN[0], un, un_length*sizeof(double));

		CL =  class_assign(c, c_length, UN, *r_UN_rows, ccount);
	}
	else if (strategy == ONE_AGAINST_ONE) {
		int nccount = (ccount*ccount - ccount)/2;
		*r_UN_rows = 2;

		UN = m_create (*r_UN_rows, nccount);

		int pos = 0;
		for (i=0; i < ccount; i++) {
			for (j=i+1; j < ccount; j++) {
				UN[0][pos] = un[i];
				UN[1][pos] = un[j];
				pos++;
			}
		}
		ccount = nccount;

		CL =  class_assign(c, c_length, UN, *r_UN_rows, ccount);
	}
	else {
		PRINTF ("Unkown multi class strategy: %i.\n", strategy);
		ret = 0;
	}

	// check for each two class problem if C is in the correct range for the nu-SVM case
	for (i=0; i < ccount; i++) {

		config->set_classifier(i);

		if (config->mode() == 1) {

			int cl1_count = find_eq(itemp, CL[i], 1, c_length);
			int cl2_count = find_eq(itemp, CL[i], -1, c_length);

			int min_count = 2 * ((cl1_count > cl2_count) ? cl2_count : cl1_count);

			if (config->C() > min_count) {
				PRINTF("C value has to be in [1 , %i].\n (class %f)", min_count, un[i]);
				ret = 0;
			}
		}
	}

	free_array(itemp);
	free_array(un);

	if (!ret) { // an error occured
		if (CL != NULL) m_free(CL);
		if (UN != NULL) m_free(UN);
	}
	else { // ok
		ret = ccount;
		*r_CL = CL;
		*r_UN = UN;
	}

	return ret;
}

// svm_wrapper trains a multi class classifier by performing training of a number of two-class classifiers
int svm_wrapper (double*** r_ALPHA, double** r_B, SubsetClassInfoDataDesc** r_SVs, double*** r_UCL, int* UCL_rows, int* ccount, 
		 ClassInfoDataDesc* X,  SVMconfig* config, Kernel* K, HashConfig* hconfig, InfoConfig* iconfig, 
		 int* learn_index, int learn_index_len, char* state_file) {

	int X_rows    = X->ssize();
	int X_columns = X->size();

	double* cl = double_array(X_columns);
	X->getClassInfo(cl);

	double** CL  = NULL;
	double** UCL = NULL;
	*ccount      = check_class_information (&CL, &UCL, UCL_rows, cl, X_columns, config);

	if (config->do_output()) PRINTF("Kernel type: %i, Number of classifiers: %i\n", K->type(), *ccount);
	if(!(*ccount)) return 0;

	//allocate the weight matrix for the support vectors
	double** ALPHA = m_create(*ccount, X_columns);
	m_zero(ALPHA, *ccount, X_columns);

	double* B = double_array(*ccount);
	memset(B, 0, *ccount*sizeof(double));

	int i=0;
	int j=0;
	int ret = 1;

	double** DECISION = NULL;
	if ((*r_ALPHA != NULL) && (*r_SVs != NULL)) { // training continued 
		if (config->do_output()) PRINTF("training continued ...\n");

		m_copy(ALPHA, *r_ALPHA, *ccount, X_columns);

		double** tDECISION = NULL;

		svm_classify (&tDECISION, NULL, ALPHA, B, *r_SVs, UCL, *UCL_rows, *ccount, K, X ,config->do_output()); 

		DECISION = m_transpose(tDECISION, X_columns, *ccount);
		m_free(tDECISION);

		// unmerge weights and class information
		for (i=0; i < *ccount; i++)  multiply (ALPHA[i], ALPHA[i], CL[i], X_columns);
	}

	tCache* G_cache = NULL;
	tCache* S_cache = NULL;

	int k_start = 0; // k_start counts the number of all training iterations over all two-class classifiers. It is required
			 // by the cache system to decide if a certain cache entry may be removed

	// I contains the incdices of those training samples that are of interest for the current two-class problem
	int* I = int_array(X_columns); 

	for (i=0; i < *ccount; i++) { // do what follows for each two class problem that arises from the multi class problem

		K->set_classifier(i); // notify the kernel about the current classifier
		config->set_classifier(i); // switch to the current classifier configuration
		iconfig->set_classifier(i); // use the information configuration setting of the current classifier
		iconfig->Init(X_columns, *ccount, config->k_max());

		// find the indices of those training samples that are of interest for the current two class problem
		int cur_X_columns = find_neq(I, CL[i], 0, X_columns);

		get_i(cl, CL[i], I, cur_X_columns);

		// allocate memory to store the langrangian alpha values for those training samples
		double* alpha = double_array(cur_X_columns); 
		get_i(alpha, ALPHA[i], I, cur_X_columns); // copy pre initialized alpha values if available

		// create subset data descriptor with respect to the entire training set
		SubsetClassInfoDataDesc* cur_X = new SubsetClassInfoDataDesc(I, cur_X_columns, X);
	
		double* decision = NULL;
		if (DECISION != NULL) {
			decision = double_array(cur_X_columns);
			get_i(decision, DECISION[i], I, cur_X_columns);
		}

		// train two class SVM
		double b=0;
		if (svm (alpha, &b, decision, &G_cache, &S_cache, cur_X, cl, k_start, *ccount, learn_index, learn_index_len, config, K, hconfig , iconfig)) {

			if ((config->mc_strategy() == ONE_AGAINST_ONE) || (*ccount == 1)) { 
				// in the one against one case or if we have only 1 classifier all values in cache can be freed 
				// otherwise the cache may be reused to train the remaining classifiers

				if (S_cache != NULL) {
					if ((config->S_size_MB() != 0) && config->do_output()) PRINTF ("sum cache stats: [misses : %i, collisions : %i, used : %i]\n", 
						S_cache->misses, S_cache->collisions, S_cache->used);

					free_cache (S_cache);
					free_array (S_cache);
					S_cache = NULL;
				}
				if (G_cache != NULL) {

					if ((config->G_size_MB() != 0) && config->do_output()) PRINTF ("kernel cache stats: [misses : %i, collisions : %i, used : %i]\n", 
						G_cache->misses, G_cache->collisions, G_cache->used);

					free_cache (G_cache);
					free_array (G_cache);
					G_cache = NULL;
				}
			}

			memset (ALPHA[i], 0, X_columns*sizeof(double));
			set_i (ALPHA[i], alpha, I, cur_X_columns); // map position of alpha values back to entire training set

			B[i] = b;
		}
		else {
			ret = 0;
			break;
		}

		if (G_cache != NULL) {
			free_cache (G_cache);
			free_array (G_cache);
			G_cache = NULL;
		}

		if (S_cache != NULL) {
			free_cache (S_cache);
			free_array (S_cache);
			S_cache = NULL;
		}

		if(state_file != NULL) {
			FILE* state_fid = fopen(state_file, "w+");
			fprintf(state_fid, "%i", i);
			fclose(state_fid);
		}

		free_array(alpha);

		if (decision != NULL) free_array(decision);

		delete cur_X;
		
		k_start += config->k_max();
	}

	double* temp = double_array(X_columns);
	int* SI = int_array(X_columns);

	// determine the support vectors
	memset(temp,0, sizeof(double)*X_columns);
	for (i=0; i < *ccount; i++)  add(temp,temp, ALPHA[i], X_columns);

	// return only weights that do not equal zero
	int SI_length = find_g (SI, temp, 0, X_columns);

	// merge weights and class information
	for (i=0; i < *ccount; i++)  multiply (ALPHA[i], ALPHA[i], CL[i], X_columns);
	
	// extract the support vector weights for each class
	double** RETURN_ALPHA = m_create(*ccount, SI_length);
	for (i=0; i < *ccount; i++) get_i(RETURN_ALPHA[i], ALPHA[i], SI, SI_length);

	if (r_SVs != NULL) *r_SVs = new SubsetClassInfoDataDesc (SI, SI_length, X);

	// return results
	*r_ALPHA = RETURN_ALPHA;
	*r_B     = B;
	*r_UCL   = UCL;

	//free temporary variables

	free_array(I);
	free_array(temp);
	free_array(SI);

	free_array(cl);

	m_free(ALPHA);
	m_free(CL);
	if(DECISION != NULL) m_free(DECISION);

	return ret;
}

// svm training using the dmmo algorithm 
//
int svm (double* r_alpha, double* r_b, double* decision,
	 tCache** r_G_cache, tCache** r_S_cache, 
	 ClassInfoDataDesc*X, double* c, int k_start, int ccount,int* learn_index, int learn_index_len,
	 SVMconfig* config, Kernel* K, HashConfig* hconfig, InfoConfig* iconfig) {

	int X_columns = X->size();
	int X_rows    = X->ssize();

	int C     = config->C();
	double C2 = config->C2(); 
//	PRINTF("C: %f\n", C2);
	
	int k_max = config->k_max();
	double epsilon = config->epsilon();
	int stop_criterion = (int) config->stop_criterion();

	int S_size_MB = config->S_size_MB();

	tCache G_cache;

	if (*r_G_cache == NULL) { // kernel cache initialization

		// theoretical number of cache rows for the kernel cache
		int G_size = X_columns;

		// G_max_usage is the number of kernel cache rows actually used
		int G_max_usage = (config->G_size_MB() * 1024*1024) / (X_columns*sizeof(double) + sizeof(int));

		if (config->do_output()) { 
			PRINTF("G_max_usage : %i\n", G_max_usage);
			FLUSH;
		}

		// determine the amount of memory that can be used for the sum cache instead 
		int unused_MB = config->G_size_MB() - (G_max_usage * (X_columns*sizeof(double) + sizeof(int)))/(1024*1024);
		unused_MB = (unused_MB < 0) ? 0 : unused_MB;

		if (S_size_MB != 0) S_size_MB = S_size_MB + unused_MB;

		int G_hash_func_params = 1;
		G_cache.delete_hash_func_params = 0;

		init_cache (&G_cache, G_size, G_max_usage, X_columns, 1, ccount * k_max, &G_hash_func_params);

		if (config->do_output()) PRINTF("G_cache: init_cache: size: %i\n", G_size);

		if ((config->G_size_MB() != 0) && config->do_output()) PRINTF ("kernel cache: [hash size : %i, cache lines : %i, data size: %i, KERNEL_CACHE_MB: %i]\n", 
						G_cache.size, G_cache.max_usage, X_columns, config->G_size_MB());
		*r_G_cache = (tCache*) char_array(sizeof(tCache));
	}
	else G_cache = **r_G_cache;


	tCache S_cache;

	if (*r_S_cache == NULL) { // sum cache initialization

		check_hash_func_params (X_columns, k_max, C, &(hconfig->hash_func_params), &(hconfig->hash_func_params_length), &S_cache.delete_hash_func_params);
	
		// cache_size determines the theoretical number of cache lines that can be used
		int S_cache_size = calc_cache_size(X_columns, k_max, C);

		// determines the maximum number of cache rows that can be allocated theoretically
		int S_cache_max_usage = (S_size_MB*1024*1024) / (C*sizeof(int) + X_columns*sizeof(double));

		if (S_size_MB == 0) S_cache_size = 0;

		init_cache (&S_cache, S_cache_size, S_cache_max_usage, X_columns, C, ccount * k_max, hconfig->hash_func_params);

		if (config->do_output()) PRINTF("S_cache: init_cache: size: %i\n", S_cache_size);

		if ((S_size_MB != 0) && config->do_output()) PRINTF ("sum cache: [hash size : %i, cache lines : %i]\n" , S_cache.size, S_cache.max_usage);

		*r_S_cache = (tCache*) char_array(sizeof(tCache));
		memset(*r_S_cache, 0, sizeof(tCache));
	}
	else S_cache = **r_S_cache;

	//------------------------------------------------
	// multiple use temporary variables

	double* temp     = double_array(X_columns);
	double* temp2    = double_array(X_columns);
	int* itemp       = int_array(X_columns);
	int itemp_length = 0;

	int* iadd = int_array (C);
	int* isub = int_array (C);

	// used during the sorting of the index values
	int bucket_size = X_columns;
	int* bucket = int_array (bucket_size);

	// t_min contains the C maximum decision values
	double* t_min = double_array (C);

	// t_max contains the C maximum decision values
	double* t_max = double_array (C);

	// Y_max contains the indices of the C maximum decision values
	int* Y_max = int_array(C);

	// SI contains the indices of the support vectors
	int* SI = int_array (X_columns);
	int SI_length = 0;

	// I contains the indices of the training data members with the C smallest decision values
	int* I = int_array(C);
	int I_length = 0;

	// J contains the indices of the training data members with the C maximum decision values
	int* J = int_array(C);
	int J_length = 0;

	// s contains the sum of the kernel function values of the support vectors for each member of the training data
	double* s = double_array (X_columns);

	// t contains s .* c
	double* t = double_array (X_columns);

	// find the C training data members with maxmimum norm
	int i=0; // calculate scalar product in kernel space for each data sample

	for (i=0; i < X_columns; i++) K->eval(temp, X, X, &i, 1, i); 

	// If the 2-norm error model is used switch to the modified kernel space
	if (config->mode() == 2) sadd(temp, temp, 1.0/C2, X_columns);

	squarert (temp, temp, X_columns); 
	smultiply(temp, temp, -1, X_columns);

	J_length = kmin (t_max, J, temp, X_columns, C);

	// determine the maximum norm of the training data in kernel space
	double R = -mean (t_max, J_length); 

	// initialize weight vector with zero
	double* alpha = r_alpha;
	double wn     = 0;

	// set s to zero
	assign (s, 0, X_columns);

	if (decision != NULL) {
		memcpy(s,decision, X_columns*sizeof(double));
		SI_length = find_g (SI, alpha, 0, X_columns);
	}
	else memset(alpha, 0, X_columns*sizeof(double));

	int k     = 0;
	int quit  = 0;

	int infop = 0;

	int* IP = int_array(X_columns);	// IP contains the indizes of class 1
	int* IN = int_array(X_columns); // IN contains the indizes of class -1

	int* IP_learn = int_array(X_columns); // IP contains the indizes of class 1 allowed to be learned 
	int* IN_learn = int_array(X_columns); // IN contains the indizes of class -1 allowed to be learned

	int IP_length = find_eq (IP, c,  1, X_columns); // find the members of class 1
	int IN_length = find_eq (IN, c, -1, X_columns); // find the members of class -1

	int IP_learn_length = 0;
	int IN_learn_length = 0;

	if (learn_index != NULL) {
		intersect (IP, learn_index, IP_length, learn_index_len, IP_learn, &IP_learn_length);
		intersect (IN, learn_index, IN_length, learn_index_len, IN_learn, &IN_learn_length);
	}
	else {
		IP_learn_length = IP_length;
		IN_learn_length = IN_length;
		memcpy(IP_learn, IP, IP_length*sizeof(int));
		memcpy(IN_learn, IN, IN_length*sizeof(int));
	}

	double gap = 0.0;
	int gap_interval=1000;
	for (k = k_start; (k < k_start + k_max) && (!quit); k++) {
	
		if ((k % gap_interval == 0)  && (k > 0)) { 

			gap_interval *= 2; // Evaluating the duality-gap is expensive, therefore the evaluation interval is doubled each time the duality-gap is evaluated
//			PRINTF("gap_interval: %i\n", gap_interval);

			if (config->do_output()) {
				PRINTF ("."); FLUSH;
				if (k % 10000 == 0)  { 
					PRINTF ("|");
					FLUSH;
				}
			}

			if (stop_criterion) {
				gap = duality_gap (C, C2, K, &G_cache, t, s, IP, IP_length, IN, IN_length, X, c, alpha, X_columns);
				quit = (gap < epsilon);
			}

		}

		// Computation of:
		// y_i * sum_j(y_j*alpha_j*K(x_j, x_i))
		if (!C2) {
			// compute decision values
			multiply (t, c, s, X_columns);
		}
		else {
			// compute decision values for 2-norm slack model (modification of the kernel)
			smultiply(temp, alpha, 1.0/C2, X_columns);
			multiply (temp, temp, c, X_columns);

			add (temp, s, temp, X_columns);
			multiply (t, c, temp, X_columns);
		}

		// learning for class 1
		learn (alpha, s, t, X, c, k, k_start,
   	  	      K, &S_cache, &G_cache, IP, IP_length,
		      t_min, t_max, Y_max, SI, IP_learn, IP_learn_length, I, J,
		      temp, itemp, iadd, isub,bucket, bucket_size, config);

		// learning for class -1
		learn (alpha, s, t, X, c, k, k_start,
   		      K, &S_cache, &G_cache, IN, IN_length,
		      t_min, t_max, Y_max, SI, IN_learn, IN_learn_length, I, J,
		      temp, itemp, iadd, isub,bucket, bucket_size, config);

	}
	if (config->do_output()) PRINTF (" Iterations : %i, Gap: %g \n", k - k_start, gap);

	//------------- normalize data --------------

	multiply (temp, s, c, X_columns); 
	multiply (temp, temp, alpha, X_columns);

	//multiply (temp, t, alpha, X_columns);

	itemp_length = find_l (itemp, alpha, 1.0/((double)C), X_columns);
	assign_i(temp, itemp, 0, itemp_length);
	wn = sqrt(sum(temp, X_columns));
	//PRINTF("wn: %g\n", wn);
	//wn = 1.0;

	//---------------------------------------------------
	// calculate threshold
	double xmin_p = 0;
	double xmax_p = 0;

	double xmin_n = 0;
	double xmax_n = 0;
	
	find_cand (alpha, t, C, IP, IP_length,
		   &xmin_p, &xmax_p, I, &I_length, J, &J_length, 
		   SI, IP_learn, IP_learn_length, t_min, t_max, Y_max, temp, itemp, bucket, bucket_size);

	find_cand (alpha, t, C, IN, IN_length,
		   &xmin_n, &xmax_n, I, &I_length, J, &J_length, 
		   SI, IN_learn, IN_learn_length, t_min, t_max, Y_max, temp, itemp, bucket, bucket_size);

	//	*r_b = ((xmin_n + xmin_p)/2)/wn; // this is how the threshold is computed according to the DMMO paper
						   // In the paper, we consider: w*x - b
						   // In the software package, we consider w*x + b

		*r_b = -((xmin_p - xmin_n)/2)/wn;  // <- which corresponds to this formula for b
						   // Note: x_min IS y_i*w*x_i (see find_cand, calc_kernel_values)

	sdivide(alpha, alpha, wn, X_columns); // normalize alpha values

	//free temporary variables
	free_array (t);
	free_array (temp);
	free_array (temp2);
	free_array (itemp);

	free_array (iadd);
	free_array (isub);

	free_array (s);

	free_array(t_min);

	free_array(t_max);
	free_array(Y_max);

	free_array(SI);

	free_array (I);
	free_array (J);

	free_array (bucket);
	free_array (IP);
	free_array (IN);

	free_array(IP_learn);
	free_array(IN_learn);

	**r_G_cache = G_cache;
	**r_S_cache = S_cache;
	
	return 1;
}
