/*
svm learning utilities written by Kai Labusch
Copyright (C) 2004-2007 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/

#include "utils.h"
#include "cache.h"
#include "min.h"
#include "svm.h"
#include "svm_params.h"
#include "svm_utils.h"
#include "kernel_func.h"

#include "DiskUtils.h"

#include <cmath>
#include <cstring>
#include <ctime>
#include <cstdlib>
#include <unistd.h>
#include <cstdio>

using namespace std;

SVMconfig::SVMconfig(void) : m_ident(NULL) , m_ident_str(NULL), pmap(NULL), nconfigs(1), output(1) {
	Default();

	DescInit();
}

SVMconfig::SVMconfig(double* l_params, int l_params_length, int tr_size) : m_ident(NULL) , m_ident_str(NULL), pmap(NULL), nconfigs(1), output(1) {
	Default();

	eval(l_params, l_params_length, tr_size);

	DescInit();
}

SVMconfig::SVMconfig(const SVMconfig* c) : nconfigs(c->nconfigs), m_ident(NULL) , m_ident_str(NULL), pmap(NULL), output(1)  {

	Default();

	memcpy(m_k_max, c->m_k_max, nconfigs*sizeof(double));
	memcpy(m_stop_criterion, c->m_stop_criterion, nconfigs*sizeof(double));
	memcpy(m_epsilon, c->m_epsilon, nconfigs*sizeof(double));
	memcpy(m_kernel_type, c->m_kernel_type, nconfigs*sizeof(double));
	memcpy(m_mode, c->m_mode, nconfigs*sizeof(double));
	memcpy(m_C2, c->m_C2, nconfigs*sizeof(double));
	memcpy(m_C, c->m_C, nconfigs*sizeof(double));
	memcpy(m_forget_sv, c->m_forget_sv, nconfigs*sizeof(double));
	memcpy(m_mc_strategy, c->m_mc_strategy, nconfigs*sizeof(double));
	memcpy(m_G_size_MB, c->m_G_size_MB, nconfigs*sizeof(double));
	memcpy(m_S_size_MB, c->m_S_size_MB, nconfigs*sizeof(double));
	memcpy(m_info_mode, c->m_info_mode, nconfigs*sizeof(double));

	DescInit();

	set_classifier(c->cur_classifier);
}

SVMconfig::~SVMconfig(void) {
	int i=0;

	for (i=0; i < PARAM_COUNT; i++) free_array(m_ident[i]);
	free_array(m_ident);

	free_array(m_ident_str);

	free_array(pmap);

	free_array(m_k_max);
	free_array(m_stop_criterion);
	free_array(m_epsilon);
	free_array(m_kernel_type);
	free_array(m_mode);
	free_array(m_C2);
	free_array(m_C);
	free_array(m_forget_sv);
	free_array(m_mc_strategy);
	free_array(m_G_size_MB);
	free_array(m_S_size_MB);
	free_array(m_info_mode);
}

void SVMconfig::Default(void) {

	cur_classifier = 0;

	m_k_max          = double_array(nconfigs);
	m_stop_criterion = double_array(nconfigs);
	m_epsilon        = double_array(nconfigs);
	m_kernel_type    = double_array(nconfigs);
	m_mode           = double_array(nconfigs);
	m_C2             = double_array(nconfigs);
	m_C              = double_array(nconfigs);
	m_forget_sv      = double_array(nconfigs);
	m_mc_strategy    = double_array(nconfigs);
	m_G_size_MB      = double_array(nconfigs);
	m_S_size_MB      = double_array(nconfigs);
	m_info_mode 	 = double_array(nconfigs);

	assign (m_k_max         , 257000         , nconfigs); // set everything to default
	assign (m_stop_criterion, 1              , nconfigs);
	assign (m_epsilon       , 1e-2           , nconfigs);
	assign (m_kernel_type   , KERNEL_RBF     , nconfigs);
	assign (m_mode          , 2              , nconfigs);
	assign (m_C2            , 1000*1000      , nconfigs);
	assign (m_C             , 1              , nconfigs);
	assign (m_forget_sv     , 1              , nconfigs);
	assign (m_mc_strategy   , ONE_AGAINST_ONE, nconfigs);
	assign (m_G_size_MB     , 100            , nconfigs);
	assign (m_S_size_MB     , 0              , nconfigs);
	assign (m_info_mode     , NO_INFO        , nconfigs);

	if (pmap == NULL) {
		pmap = (double**) pointer_array(PARAM_COUNT);

		set_pointers();
	}
}

void SVMconfig::set_pointers(void) {
	pmap[K_MAX_POS]                = m_k_max       + cur_classifier;
	pmap[STOP_CRITERION_POS]       = m_stop_criterion + cur_classifier;
	pmap[EPSILON_POS]              = m_epsilon     + cur_classifier;
	pmap[KERNEL_TYPE_POS]          = m_kernel_type + cur_classifier;
	pmap[MODE_POS]                 = m_mode        + cur_classifier;
	pmap[C_POS]                    = m_C           + cur_classifier;
	pmap[FORGET_SV_POS]            = m_forget_sv   + cur_classifier;
	pmap[MULTI_CLASS_STRATEGY_POS] = m_mc_strategy + cur_classifier;
	pmap[INFO_MODE_POS]            = m_info_mode   + cur_classifier;
	pmap[G_CACHE_MB_POS]           = m_G_size_MB   + cur_classifier;	
	pmap[S_CACHE_MB_POS]           = m_S_size_MB   + cur_classifier;
}

void SVMconfig::DescInit(void) {

	int i=0;

	if(m_ident == NULL) {
		m_ident = (char**) pointer_array(PARAM_COUNT);
		for (i=0;i < PARAM_COUNT; i++) m_ident[i] = char_array(255);
	}

	sprintf (m_ident[K_MAX_POS], "k_max");
	sprintf (m_ident[STOP_CRITERION_POS], "stop_criterion");
	sprintf (m_ident[EPSILON_POS], "epsilon");
	sprintf (m_ident[KERNEL_TYPE_POS],"kernel_type");
	sprintf (m_ident[MODE_POS],"svm_type");
	sprintf (m_ident[C_POS],"C");
	sprintf (m_ident[FORGET_SV_POS],"forget_sv");
	sprintf (m_ident[MULTI_CLASS_STRATEGY_POS], "mc_strategy");
	sprintf (m_ident[INFO_MODE_POS],"info_mode");
	sprintf (m_ident[G_CACHE_MB_POS], "KERNEL_CACHE_MB");
	sprintf (m_ident[S_CACHE_MB_POS], "SUM_CACHE_MB");

	if(m_ident_str == NULL) m_ident_str = char_array(PARAM_COUNT*255);

	sprintf (m_ident_str,"[");
	const char* delim = "";

	for (i=0; i < PARAM_COUNT; i++) {
		sprintf (m_ident_str + strlen(m_ident_str), "%s%s: %g", delim, m_ident[i], *(pmap[i]));
		delim = ",";
	}

	sprintf (m_ident_str + strlen(m_ident_str), "]");
}

void SVMconfig::set(const char* name, double val, int tr_size) {
	int* pos = int_array(PARAM_COUNT);
	int i=0;

	int count = sfind_eq(pos, m_ident, name, PARAM_COUNT);

	if (count) {
		double* temp = double_array(PARAM_COUNT);
		get(temp, PARAM_COUNT);

		for (i=0; i < count; i++) {
			if (output) PRINTF("SVMconf: \"%s\" =  %g\n", name, val);
			temp[pos[i]] = val;
		}

		int output=0;
		eval(temp, PARAM_COUNT, tr_size, output);
		free_array(temp);
	}
	//else PRINTF("Warning! Could not find parameter \"%s\" (SVMconfig::set)\n", name);

	free_array(pos);
}

int SVMconfig::get(double* l_params, int l_params_length) {

	if (l_params_length != PARAM_COUNT) {
		PRINTF ("Wrong learning parameters.\n");
		return 0;
	}
	
	int i=0;
	for (i=0; i < l_params_length;  i++) l_params[i] = *(pmap[i]);

	if (*m_mode == 1)      l_params[C_POS] = *m_C;
	else if (*m_mode == 2) l_params[C_POS] = *m_C2;
        else {
		PRINTF ("\nUnknown SVM mode %i.\n", ((int) *m_mode));
		return 0;
	}

	return 1;
}


int SVMconfig::eval(double* l_params, int l_params_length, int tr_size, int output) {

	if (l_params_length > PARAM_COUNT) {
		PRINTF ("Wrong learning parameters.\n");
		return 0;
	}
	
	int i=0;
	for (i=0; i < l_params_length;  i++) *(pmap[i]) = l_params[i];

	double P = *m_C;

	if (output) PRINTF ("k_max: %i, ", (int) *m_k_max);

	if ((*m_stop_criterion) && output)  PRINTF (" stop_criterion: %g, epsilon : %g,", *m_stop_criterion, *m_epsilon);

	if (*m_mode == 1) { // nu-SVM mode, soft margin is realized by averaging over several data points
		*m_C2 = 0;
		*m_C  = P;

		if (output) PRINTF (" nu : %f,", (2.0* *m_C)/((double)tr_size));
	}
	else if (*m_mode == 2) { // C2-SVM mode, soft margin is realized by modifying the kernel function
		*m_C2 = P;
		*m_C = 1;
	
		if (output) PRINTF (" C : %f,", *m_C2);
	}
        else {
		PRINTF ("\nUnknown SVM mode %i.\n", ((int) *m_mode));
		return 0;
	}

	if (output) PRINTF (" forgetting mode : %i,", ((int) *m_forget_sv));
	if (output) PRINTF (" mc strategy : %i,"    , ((int) *m_mc_strategy));
	if (output) PRINTF (" info_mode : %i\n"     , ((int) *m_info_mode));

	FLUSH;

	return 1;
}

void SVMconfig::set_classifier(int i){
	if (i < nconfigs) cur_classifier = i;
//	else PRINTF("Cannot use configuration %i. Using configuration %i ...\n", i, cur_classifier);
	set_pointers();
	DescInit();
}

int SVMconfig::kernel_type(void) {
	return (int) m_kernel_type[cur_classifier];
}

int SVMconfig::k_max(void) {
	return (int) m_k_max[cur_classifier];
}

int SVMconfig::mode(void) {
	return (int) m_mode[cur_classifier];
}

int SVMconfig::C(void) {
	return (int) m_C[cur_classifier];
}

double SVMconfig::C2(void) {
	return m_C2[cur_classifier];
}

double SVMconfig::epsilon(void) {
	return m_epsilon[cur_classifier];
}

double SVMconfig::stop_criterion(void) {
	return m_stop_criterion[cur_classifier];
}


int SVMconfig::forget_sv(void) {
	return (int) m_forget_sv[cur_classifier];
}

int SVMconfig::mc_strategy(void) {
	return (int) m_mc_strategy[cur_classifier];
}

int SVMconfig::info_mode(void) {
	return (int) m_info_mode[cur_classifier];
}

int SVMconfig::G_size_MB(void) {
	return (int) m_G_size_MB[cur_classifier];
}

int SVMconfig::S_size_MB(void) {
	return (int) m_S_size_MB[cur_classifier];
}

char** SVMconfig::ident(void) {
	return m_ident;
}

char*  SVMconfig::ident_str(void) {

	ident_str(m_ident_str);

	return m_ident_str;
}

void SVMconfig::ident_str(char* dest) {

	sprintf (m_ident_str,"[");
	const char* delim = "";

	for (int i=0; i < PARAM_COUNT; i++) {
		sprintf (m_ident_str + strlen(m_ident_str), "%s%s: %g", delim, m_ident[i], *(pmap[i]));
		delim = ",";
	}

	sprintf (m_ident_str + strlen(m_ident_str), "]");

	if (dest != m_ident_str) sprintf(dest, "%s", m_ident_str);
}

int SVMconfig::do_output(void) {
	return output;
}

void SVMconfig::output_on(void) {
	output=1;
}

void SVMconfig::output_off(void) {
	output=0;
}


InfoConfig::InfoConfig(int imode, double* iparams, int iparams_len) : m_info_mode(imode) , r_info(NULL), info_rows(NULL), info_columns(NULL), 
								      info_params(iparams), info_params_length(iparams_len), cur_classifier(0), m_ccount(0), position(0) {
}

InfoConfig::InfoConfig(const InfoConfig* c) : m_info_mode(c->m_info_mode) , r_info(NULL), info_rows(NULL), info_columns(NULL), 
					       info_params(c->info_params), info_params_length(c->info_params_length), cur_classifier(c->cur_classifier), m_ccount(0), position(0) {
}

InfoConfig::~InfoConfig(void) {

	if (r_info != NULL) {
		int i;
	
		for (i=0; i < m_ccount; i++)
			if (r_info[i] != NULL) m_free(r_info[i]);	

		free_array(r_info);

		free_array(info_rows);
		free_array(info_columns);
	}
}

void InfoConfig::Init(int num_samples, int ccount, int k_max) {

	if(r_info == NULL) {
		m_ccount = ccount;	
	
		r_info = (double***) pointer_array(ccount);
		memset(r_info, 0, ccount * sizeof(double**));

		info_rows    = int_array(ccount);
		info_columns = int_array(ccount);
	}

	info_rows[cur_classifier]    = 0;
	info_columns[cur_classifier] = 0;

	switch (m_info_mode) {
		case NO_INFO: break;
		case ZERO_SET_INFO:
			info_rows   [cur_classifier] = num_samples;
			info_columns[cur_classifier] = 2;
		break;
		case CONV_INFO:
			info_rows[cur_classifier] = num_samples;
			if(info_params == NULL)  info_columns[cur_classifier] = k_max;
			else info_columns[cur_classifier] = info_params_length;
		break;
		default: 
			PRINTF("Unknown info mode: %i\n", m_info_mode);
		break;
	}

	if (info_rows[cur_classifier] * info_columns[cur_classifier] > 0) {
		r_info[cur_classifier] = m_create(info_columns[cur_classifier], info_rows[cur_classifier]);
		m_zero(r_info[cur_classifier], info_columns[cur_classifier], info_rows[cur_classifier]);
	}

}

void InfoConfig::store_info(double* vals, int len, int k) {

	if ((m_info_mode == CONV_INFO)) {
		if (info_params == NULL) memcpy(r_info[cur_classifier][position++], vals, len*sizeof(double));
		else if(k == info_params[position])  memcpy(r_info[cur_classifier][position++], vals, len*sizeof(double));
	}

}

void InfoConfig::set_classifier(int i) {
	cur_classifier = i;
	position = 0;
}

double** InfoConfig::cur_r_info(void) {
	return r_info[cur_classifier];
}

int InfoConfig::cur_rows(void) {
	return info_rows[cur_classifier];
}

int InfoConfig::cur_columns(void) {
	return info_columns[cur_classifier];
}


int InfoConfig::info_mode(void) {
	return m_info_mode;
}


HashConfig::HashConfig(int* hparams, int hparam_len) : hash_func_params(hparams), hash_func_params_length(hparam_len), allocated(0) {
} 

HashConfig::~HashConfig(void) {
}

