/*
(Soft)MaxMinOver learning algorithm written by Kai Labusch
Copyright (C) 2004 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/

#include "utils.h"
#include "cache.h"
#include "min.h"
#include "svm.h"
#include "kernel_func.h"
#include "svm_params.h"
#include "svm_utils.h"
#include "Data.h"


#include "DiskUtils.h"

#include <cmath>
#include <cstring>
#include <ctime>
#include <cstdlib>
#include <unistd.h>
#include <cstdio>

using namespace std;

void learn (double* alpha, double* s, double* t, ClassInfoDataDesc* X, double* c, int k, int k_start, 
	   Kernel* K, tCache* S_cache, tCache* G_cache, int* CLI, int CLI_length,
	   double* t_min, double* t_max, int* Y_max, int* SI, int* LI, int LI_length, int* I, int* J,
	   double* temp, int* itemp, int* iadd, int* isub,
	   int* bucket, int bucket_size, SVMconfig* config) {

	int X_rows    = X->ssize();
	int X_columns = X->size();
	
	int mode      = config->mode();
	int C         = config->C();
	double C2     = config->C2();
	double epsilon = config->epsilon();

	double l = 0;

	int I_length=0;
	int J_length=0;

	int temp_length = 0;
	int itemp_length = 0;

	int iadd_length = 0;
	int isub_length = 0;

	double xmin = 0;
	double xmax = 0;

	double learn = 0;

	find_cand (alpha, t, C, CLI, CLI_length,
		   &xmin, &xmax, I, &I_length, J, &J_length, 
		   SI, LI, LI_length, t_min, t_max, Y_max, temp, itemp, bucket, bucket_size);
		
	if (config->forget_sv()) {

		learn = 2.0/((double)C);
	
		// forget the support vectors with maximum decision value
		ssub_i(alpha, alpha, 1.0/((double)C), J, J_length);

		// update the sum of the kernel values
		// check if the sum-cache contains what we need
		double* s2 = NULL;
		int* key   = J;
		int ret    = 0;

		if ((ret = get_cache_entry (&s2, &key, k, S_cache)) <= 0) { // if S_cache does not contain what we need

			// determine how s2 can be modified to get what we need
			if (ret == 0) calc_sum_diff (J, key, J_length, iadd, isub, &iadd_length, &isub_length);

			if ((iadd_length + isub_length < C) && (ret == 0)) { // if modifying the cache entry is more efficient
				// sum-cache does not contain what we need, therefore compute the sum
				calc_kernel_values (s2, iadd, iadd_length, isub, isub_length, X, c, K , k, G_cache, itemp);
			} // otherwise compute the entire sum
			else  calc_kernel_values (s2, J, J_length, NULL, 0, X, c, K , k, G_cache, itemp);

			memcpy (key, J, J_length*sizeof(int));
		}

		if (C != 1) {
			smultiply(temp,s2, 1/((double) C), X_columns);
			sub (s, s, temp, X_columns);
		}
		else sub (s, s, s2, X_columns);

	}		
	else learn = 1.0/((double)C);

	// update the sum of the kernel values of all training data members
	double* s2 = NULL;
	int* key = I;
	int ret = 0;

	if ((ret = get_cache_entry (&s2, &key, k, S_cache)) <= 0 ) {

		// determine how s2 can be modified to get what we need
		if (ret == 0) calc_sum_diff (I, key, I_length, iadd, isub, &iadd_length, &isub_length);

		if ((iadd_length + isub_length < C) && (ret == 0)) { // if modifying the cache is more efficient

			calc_kernel_values (s2, iadd, iadd_length, isub, isub_length, X, c, K , k, G_cache, itemp);
		} // else compute the entire sum
		else calc_kernel_values (s2, I, I_length, NULL, 0 , X, c, K , k, G_cache, itemp);

		memcpy (key, I, sizeof(int)* I_length);
	}

	smultiply(temp, s2 , learn, X_columns);	
	add (s, s, temp, X_columns);

	// update weights
	sadd_i(alpha, alpha, learn, I, I_length);
}

//--------------------------------------------------------------------------
// determine the training samples to be considered in the current learning step
//
// alpha : counts how often each training sample was selected for learning
// t     : distance of each training sample with respect to the current hyperplane, i.e.
//
//         y_i * sum_j(y_j*alpha_j*K(x_j, x_i)) (to be provided by the caller)
//
// t_xmin : return the C smallest distances, i.e. entrys of t
// t_xmax : return the C largest distances of training samples with (alpha_i != 0), i.e., current SVs
// r_xmin : mean of t_min
// r_xmax : mean of t_max times -1
// r_I    : return the indices of the C smallest distances that belong to r_xmin
// r_J    : return the indices of the C largest distances that belong to r_xmax
// SI     : return the indices of the current SVs
//
// Y_max, temp, itemp, bucket : temporary variables
//
void find_cand (double* alpha, double* t, int C, int* CLI, int CLI_length,
		double* r_xmin, double* r_xmax, int* r_I, int* r_I_length, int* r_J, int* r_J_length, 
		int* SI, int* LI, int LI_length, double* t_min, double* t_max, int* Y_max, double* temp, int* itemp, int* bucket, int bucket_size) {

	int t_min_length = 0;
	int t_max_length = 0;
	int itemp_length = 0;
	int SI_length    = 0;
	int temp_length  = 0;
	int Y_max_length = 0;

	// select those samples to be considered for the min step
	temp_length = get_i(temp, t, LI, LI_length);

	// find the C smallest decision values as well as their indices from the remaining set
	t_min_length = *r_I_length = kmin (t_min, r_I, temp, temp_length, C);

	// make absolute indices
	itemp_length = iget_i(itemp, LI, r_I, *r_I_length); 

	// use the cache as sparingly as possible, to accomplish this sort r_I
	*r_I_length = bucket_sort (r_I, itemp, itemp_length, bucket, bucket_size);

	// select alpha values to be considered
	temp_length = get_i(temp, alpha, CLI, CLI_length);

	// find the indices of the support vectors from the remaining set
	itemp_length = find_neq (itemp, temp, 0, temp_length);
	
	// make absolute indices
	SI_length = iget_i(SI, CLI, itemp, itemp_length);

	// get the decision values of the support vectors
	temp_length = get_i(temp, t, SI, SI_length);

	// find the C support vectors with maximum decision value
	smultiply(temp, temp, -1, temp_length);
	t_max_length = Y_max_length = kmin (t_max, Y_max, temp, temp_length, C);

	// compute mean
	*r_xmin = mean (t_min, t_min_length);

	*r_xmax = 0;
	if (t_max_length > 0) {
		*r_xmax = -mean (t_max, t_max_length);
		itemp_length = iget_i(itemp, SI, Y_max, Y_max_length);
		
		// use the cache as sparingly as possible , to accomplish this sort J
		*r_J_length = bucket_sort (r_J , itemp, itemp_length, bucket, bucket_size);
	}
	else {
		// currently there aren't any support vectors available
		*r_xmax = *r_xmin;
		*r_J_length = 0;
	}
}

void calc_sum_diff (int* dest_key, int* cur_key, int key_length, int* I_add, int* I_sub, int* I_add_length, int* I_sub_length) {
	int dest_i = 0;
	int cur_i = 0;

	int add_i = 0;
	int sub_i = 0;

	int dest = 0;
	int cur = 0;

	while ((dest_i < key_length) || (cur_i < key_length)) {

		dest = (dest_i < key_length) ? dest_key[dest_i] : dest_key[key_length - 1];
		cur = (cur_i < key_length) ? cur_key[cur_i] : cur_key[key_length -1];

		if ((dest < cur) && (dest_i < key_length)){

			I_add[add_i] = dest_key[dest_i];		

			add_i++;
			dest_i++;
		}
		else if ((cur < dest) && (cur_i < key_length)){

			I_sub[sub_i] = cur_key[cur_i];

			sub_i++;
			cur_i++;
		}
		else {
			if ((dest_i < key_length) && (cur_i < key_length)) {
				dest_i++;
				cur_i++;
			}
			else if (dest_i < key_length) {
				I_add[add_i] = dest_key[dest_i];		

				add_i++;
				dest_i++;
			}
			else if (cur_i < key_length) {
				I_sub[sub_i] = cur_key[cur_i];

				sub_i++;
				cur_i++;
			}
		} 
	}
	
	*I_add_length = add_i;
	*I_sub_length = sub_i;
}

// This function only operates on the kernel cache! For the kernel cache always cache->size = X_columns holds.
void calc_kernel_values (double* dest, int* I_add, int I_add_length, int* I_sub, int I_sub_length, 
			 ClassInfoDataDesc* X, double* c, Kernel* K, int time, tCache* G_cache, int* itemp) {

	int X_rows = X->ssize();
	int X_columns = X->size();

	int i=0;
	int j=0;
	int itemp_len=0;

	if (I_sub_length == 0) memset (dest, 0, X_columns*sizeof(double));

	for (i=0; i < I_add_length; i++) {

		double* G_entry =  NULL;
		int* r_key = I_add + i;

		// check if the kernel cache contains what we need
		if(get_cache_entry (&G_entry, &r_key, time, G_cache) <= 0) {

			// kernel cache does not contain what we need , therefore the kernel values have to be computed

			if (G_cache->cache_on) { // search for kernel values that are already present and can be reused

				itemp[0] = I_add[i]; // always compute the missing data point
				itemp_len = 1;
				for (j=0; j < G_cache->size; j++) {
					if (G_cache->Values[j] != NULL) G_entry[j] = (G_cache->Values[j])[I_add[i]];
					else if (j!=I_add[i]) {
						itemp[itemp_len] = j;
						itemp_len++;
					}
				}
				K->eval(G_entry, X, X, itemp, itemp_len, I_add[i]);
			}		
			else K->eval(G_entry, X, X, NULL, X_columns, I_add[i]);

			*r_key = I_add[i];	
		}

		if (c[I_add[i]] == 1) add(dest, dest, G_entry, X_columns); // This corresponds to the computation of
		else sub(dest, dest, G_entry, X_columns);		   // sum_j (y_j * alpha_j * K(x_j, x_i) )
	}								   // for each training sample x_i

	for (i=0; i < I_sub_length; i++) {

		double* G_entry =  NULL;
		int* r_key = I_sub + i;

		// check if the kernel cache contains what we need
		if(get_cache_entry (&G_entry, &r_key, time, G_cache) <= 0) {
			// kernel cache does not contain what we need , therefore the kernel values have to be computed

			if (G_cache->cache_on) { // search for kernel values that are already present and can be reused
				itemp[0] = I_sub[i]; // always compute the missing data point
				itemp_len = 1;

				for (j=0; j < G_cache->size; j++) {
					if (G_cache->Values[j] != NULL) G_entry[j] = (G_cache->Values[j])[I_sub[i]];
					else if (j!=I_sub[i]) {
						itemp[itemp_len] = j;
						itemp_len++;
					}
				}
				K->eval(G_entry, X, X, itemp, itemp_len, I_sub[i]);
			}
			else K->eval(G_entry, X, X, NULL, X_columns, I_sub[i]);
			
			*r_key = I_sub[i];	
		}
		if (c[I_sub[i]] == 1) sub (dest, dest, G_entry, X_columns); // This corresponds to the computation of 
		else add(dest, dest, G_entry, X_columns);		    // sum_j (y_j * alpha_j * K(x_j,x_i) )
	}								    // for each training sample x_i
}

// s : hard margin decision values
// t : soft-margin decision values x class information
double duality_gap (double C, double C2, Kernel* K, tCache* G_cache, double* t, double* s, 
		    int* IP, int IP_length, int* IN, int IN_length, 
		    ClassInfoDataDesc* X, double* c, double* alpha, int X_columns) {

	double gap = 1e10;
	int time   = -1; // use G_cache in read-only mode

	int* SVs      = int_array(X_columns); 
	double* temp  = double_array(X_columns); 
	double* temp2 = double_array(X_columns);
	int* itemp    = int_array(X_columns);

	int nSVs = find_neq(SVs, alpha, 0, X_columns);

	memset(s, 0, X_columns*sizeof(double));

	double sq_norm_hard=0;
	double sq_norm_soft=0;

	for (int i=0; i < nSVs; i++) {
		double* G_entry = temp;
		int* r_key      = SVs + i;
		int itemp_len   = 0;

		if(get_cache_entry (&G_entry, &r_key, time, G_cache) <= 0) {

			itemp[0]  = SVs[i]; // always compute the missing data point
			itemp_len = 1;
			for (int j=0; j < G_cache->size; j++) {
				if (G_cache->Values[j] != NULL) G_entry[j] = (G_cache->Values[j])[SVs[i]];
				else if (j!=SVs[i]) {
					itemp[itemp_len] = j;
					itemp_len++;
				}
			}
			K->eval(G_entry, X, X, itemp, itemp_len, SVs[i]); // G_entry = K(x_k,x_j), forall j=1,...,N and k=SVs[i]
		}

		smultiply(temp2, G_entry, alpha[SVs[i]]*c[SVs[i]], X_columns); // temp2 = alpha(k)*c(k)*G_entry
		add(s, s, temp2, X_columns);

		memcpy(temp, G_entry, X_columns*sizeof(double));
		temp[SVs[i]] += 1.0/C2;	
		smultiply (temp, temp, alpha[SVs[i]]*c[SVs[i]], X_columns);

		multiply(temp, temp, alpha, X_columns);
		multiply(temp, temp, c, X_columns);

		multiply(temp2, temp2, alpha, X_columns); // temp2 = alpha(k)*c(k)*alpha*c*G_entry
		multiply(temp2, temp2, c, X_columns);

		sq_norm_hard += sum(temp2, X_columns);
		sq_norm_soft += sum(temp, X_columns);
	}

	// Computation of:
	// y_i * sum_j(y_j*alpha_j*K(x_j, x_i))

	if (C2) {
		double* hard_margin = double_array(X_columns);
		double* xi          = double_array(X_columns);

		// compute decision values for 2-norm slack model (modification of the kernel)
		smultiply(temp, alpha, 1.0/C2, X_columns); // temp = alpha .* c /C2
		multiply (temp, temp, c, X_columns);

		add (temp, s, temp, X_columns); // temp = s + temp
		multiply (t, c, temp, X_columns);// t = (s + temp) .* c

		get_i(temp2, t, IP, IP_length);
		double t_min_p = temp2[dmin(temp2, IP_length)];
		
		get_i(temp2, t, IN, IN_length);
		double t_min_n = temp2[dmin(temp2, IN_length)];

		double b = 0.5*(t_min_p - t_min_n);

		ssub(temp, temp, b, X_columns);
		multiply(temp, c, temp, X_columns);

		double factor = temp[dmin(temp, X_columns)];

		ssub(hard_margin, s , b, X_columns);
		multiply(hard_margin, hard_margin, c, X_columns);
		sdivide(hard_margin, hard_margin, factor, X_columns);

		sq_norm_hard /= factor*factor;
		sq_norm_soft /= factor*factor;

		double dual = sum(alpha, X_columns)/factor - 0.5*sq_norm_soft;

		int num_errors = find_l(itemp, hard_margin, 1.0, X_columns);
		get_i(xi, hard_margin, itemp, num_errors); // temp = hard_margin(hard_margin < 1)

		smultiply(xi,xi, -1.0, num_errors);
		sadd(xi,xi,1.0, num_errors);

		multiply(xi, xi, xi, num_errors);

		double primal = 0.5*sq_norm_hard + 0.5*C2*sum(xi, num_errors);

		gap = (primal - dual)/(primal + 1);

		free_array(hard_margin);
		free_array(xi);
	}
	else {
		// compute decision values
		multiply (t, c, s, X_columns);
	}

	free_array(SVs);
	free_array(temp);
	free_array(temp2);
	free_array(itemp);

	return gap;
}
