function [alpha, b, SVs, cSV, ucl, sv_indizes, nf, k_params, bmerr, bm_svcount, C] = CVMaxMinOver (X, c, k, kernel_type, grid_count, precompute_kernel, max_level, KERNEL_CACHE_MB)
%
%[alpha, b, SVs, cSV, ucl, sv_indizes, nf, k_params, bmerr, bm_svcount, C] = CVMaxMinOver (X, c, k, kernel_type, grid_count, ...
%                                                                                       precompute_kernel, max_level, KERNEL_CACHE_MB, save_memory)
%
% Determine optimal SVM hyper-parameters via k-fold cross-validation. Perform grid search with respect to
% the SVM hyper-parameters. Grid search stops if the mean validation error rate could not be improved 
% or maximum number of refinement levels has been reached (max_level).
% For SVM training  the SoftDoubleMaxMinOver algorithm [1] is employed. Using the optimal hyper-parameters
% a final training on the entire data set is performed.
%
% typical usage:
%
% [alpha, b, SVs, cSV, ucl, sv_indizes, nf, k_params] = CVMaxMinOver (X, c, k, kernel_type)
%
% Hint: If you have enough memory set precompute_kernel to 1. It usually significantly increases the speed of 
%       computation.
% See also: http://www.inb.uni-luebeck.de/tools-demos/maxminover
%
%Input: 
% X : Either matrix containing the training data as column vectors or name 
%     of ascii file containing the training data (one sample per row).
% c : Either row vector with class information of training data or name of 
%     ascii file containing the class information.
% k : Perform k-fold Cross-Validation (default: 10)
% kernel_type : 0->linear 
%               1->rbf (L2 distance)
%               2->polynom
%               3->rbf (L0 distance)
% grid_count : number of grid points in each search dimension (optional, default 5)
% precompute_kernel : 0 -> do not precompute kernel matrix 1 -> precompute kernel (optional, default 0)
% max_level : maximum number of recursive refinements during grid search (optional, default 5)
% KERNEL_CACHE_MB : maximum amount of memory to be used for kernel caching (optional, default 500)
% save_memory : 0-> off, default 1-> use as few memory as possible
%
%Return Values:
%
% alpha : weights of the support vectors (for every two class problem)
% b     : threshold (for every two class problem)
% SVs : support vectors
% cSV : class information of the support vectors 
% ucl : class assignment for each two class problem, i.e., the training data is internally 
%       relabeled where for each two class problem the labels are in {-1,1}.
%       In case of one-against-one training for the i-th two-class classifier ucl(1,i) 
%       describes which class got label 1 whereas ucl(2,i) describes which class got label -1.
%       In case of one against-all training for the i-th two-class classifier ucl(i) describes 
%       which class got the label 1 whereas all other classes got the label -1.
% sv_indizes : indizes of the support vectors with regard to the entire data set
% nf       : normalization factor (to be feeded to MaxMinOverClassify and TrainMaxMinOver 
%	     if the returned parameters are to be used for these functions)
% k_params : optimal kernel parameters determined by cross-validation
% bmerr : best mean error during validation
% bm_svcount : mean number of SVs for the best hyper parameter combination during validation
% C : best softness
%
%----------
%
% Note that the method performs an internal normalization that leads to the norm factor nf!!!
% In order to obtain a reasonable result this normalization factor must be feeded to MaxMinOverClassify!!!
%
% In case of linear classification the normal vector of the hyper-plane can be obtained by:
%
%     w = sum(SVs*diag(alpha), 2);
%
% in order to obtain a valid threshold one has to consider: b = b/(nf^2);
%
%-----------
% [1] Thomas Martinetz, Kai Labusch, and Daniel Schneegaß. SoftDoubleMaxMinOver: Perceptron-like Training of Support Vector Machines. 
%     IEEE Transactions on Neural Networks, 20(7):1061-1072, 2009.
%     http://www.inb.uni-luebeck.de/publications/pdfs/MaLaSc09.pdf
%
