function [alpha, b, SVs, cSV, ucl, sv_indizes, info] = TrainMaxMinOver (X, c, l_params, k_params, nf, train_index, save_memory, init_alpha, uv_hash_params, info_params, mask)
%[alpha, b, SVs, cSV, ucl, sv_indizes, info] = TrainMaxMinOver (X, c, l_params, k_params, nf, train_index, save_memory, init_alpha, uv_hash_params, info_params, mask)
%
% Train SVM classifier with SoftDoubleMaxMinOver [1] algorithm.
% Train SVM classifier with SoftDoubleMinOver algorithm, if forget_sv is set to zero.
%
% typical usage:
%
% [alpha, b, SVs, cSV, ucl, sv_indizes] = TrainMaxMinOver (X, c, l_params, k_params);
%
% if k_params has been obtained by CVMaxMinOver:
%
% [alpha, b, SVs, cSV, ucl, sv_indizes] = TrainMaxMinOver (X, c, l_params, k_params, nf);
%
% See also: http://www.inb.uni-luebeck.de/tools-demos/maxminover
%
%Input: 
% X : Either matrix containing the training data as column vectors or name 
%     of ascii file containing the training data (one sample per row).
% c : Either row vector with class information of training data or name of 
%     ascii file containing the class information.
%
% l_params : Learning parameters ->
%
%              [k_max,kernel_type,C,svm_type,forget_sv, mc_strategy, stop_criterion,epsilon,info_mode,KERNEL_CACHE_MB,SUM_CACHE_MB]
%
% 	k_max       : maximum number of iterations if terminination criterion does not hold (use duality gap by default)
% 	kernel_type : 0 - linear, 1 - rbf, 2 - polynomial, (default 1)
%
% 	C : 	if svm_type == 1 holds, this is the  number of training samples considered together, i.e., C in [1, min(mn, mp)], 
%			where mn and mp are the number of training samples for class -1 resp. 1.
%     		if svm_type == 2 holds (default), this is the weight for sum of classification errors according to the 2-norm-slack-SVM 
%           error model, i.e., C in [0, inf] (default C=10^6 (hard-margin)).
%
% 	svm_type : error model: 1 - nu-SVM, 2 - 2-norm-slack-SVM (default 2)
%
% 	forget_sv : 	
%           0 - forgetting off 
%           1 - forgetting on (default)
%	mc_strategy :   multi class strategy:
%           0 - one against all
%           1 - one against one (default)
%
%	stop_criterion : determines if duality-gap should be used as stopping criterion. 
%           0 - do not use duality gap
%           1 - use duality gap (default)
%
% 	epsilon : lower bound of duality gap, learning stops if reached (default 10^-3)
%
%	info_mode : determines which information about the learning process should be returned (deprecated)
%
%	KERNEL_CACHE_MB : maximum amount of memory used for kernel caching, 0 - kernel cache off (default up to 100 MB)
%	SUM_CACHE_MB    : maximum amount of memory used for sum caching, 0 - sum cache off (default)
%
% 	k_params : kernel parameters: 
%		   linear     : []
%		   rbf        : [gamma] (default 1)
%		   polynomial : [degree, k] (default [1,1])
%
% nf : optional normalization factor (default: 1.0, might be obtained by CVMaxMinOver)
% train_index : indices of the subset that should be used for training (optional, use all samples by default)
% save_memory : 0 - normal training (load entire training data from file, default)
%		        1 - save memory by loading only training data that is currently used
%
% init_alpha : initial weights for entire training data. Can be used to continue learning (optional, default [])
%
% uv_hash_params : random parameters to be used for universal hashing, to be used only in nu-svm mode (optional)
%
% info_params : if info_mode is set to 3 this vector has to contain the time steps when the alpha values should be saved (optional)
%
% mask : matrix containing the indices of those dimensions to be used to train each two class classifier. Each column of this matrix
%	     defines the dimensions to be used for a two class classifier (optional) 
%
%Return Values:
%
% alpha : weights of the support vectors (for every two class problem)
% b     : threshold (for every two class problem)
% SVs : support vectors (empty matrix in case of save_memory == 1)!!!!
% cSV : class information of the support vectors 
% ucl : class assignment for each two class problem, i.e., the training data is internally 
%       relabeled where for each two class problem the labels are in {-1,1}.
%       In case of one-against-one training for the i-th two-class classifier ucl(1,i) 
%       describes which class got label 1 whereas ucl(2,i) describes which class got label -1.
%       In case of one against-all training for the i-th two-class classifier ucl(i) describes 
%       which class got the label 1 whereas all other classes got the label -1.
% sv_indizes : indizes of the support vectors with respect to the entire data set
% info : information on the learining process depending on info_mode
%
% The standard settings for l_params and k_params can be obtained as
% follows:
%
% params   = lparam_set(1);
% l_params = lparam_get('learn_params', params, 1);
% k_params = lparam_get('kernel_params', params, 1);
%
% A description of the parameters can be obtained as follows:
% lparam_get('description', params)
%
% One can set individual parameters by appending their name and value to the argument list of lparam_set:
%
%   params = lparam_set(1,'k_max', 5000); % get default parameter set of the DMMO package and alter the k_max setting
%   params = lparam_set(1,'kernel_type', 'linear'); % get default parameter set of the DMMO package, use linear kernel instead of rbf
%   params = lparam_set(1,'kernel_type', 'polynom'); % get default
%   parameter set of the DMMO package, use polynomial kernel instead of rbf
%
%-------------
% [1] Thomas Martinetz, Kai Labusch, and Daniel Schneegaß. SoftDoubleMaxMinOver: Perceptron-like Training of Support Vector Machines. 
%     IEEE Transactions on Neural Networks, 20(7):1061-1072, 2009.
%
